# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""A clock class."""

import logging
import warnings
from decimal import Decimal
from fractions import Fraction
from logging import Logger
from typing import ClassVar, Type, Union

import cocotb
from cocotb._py_compat import (
    Literal,
    TypeAlias,
    cached_property,
)
from cocotb._typing import TimeUnit
from cocotb.handle import (
    Deposit,
    Force,
    Immediate,
    LogicObject,
    _GPISetAction,
    _trust_inertial,
)
from cocotb.simulator import clock_create
from cocotb.task import Task
from cocotb.triggers import (
    Event,
    FallingEdge,
    NullTrigger,
    RisingEdge,
    Timer,
    ValueChange,
)
from cocotb.utils import get_sim_steps, get_time_from_sim_steps

__all__ = ("Clock",)

Impl: TypeAlias = Literal["gpi", "py"]


_valid_impls = ("gpi", "py")


class Clock:
    r"""Simple 50:50 duty cycle clock driver.

    .. code-block:: python

        c = Clock(dut.clk, 10, "ns")
        c.start()

    Args:
        signal: The clock pin/signal to be driven.
        period:
            The clock period.
            Must be a multiple of the time precision of the simulator.

        unit:
            One of ``'step'``, ``'fs'``, ``'ps'``, ``'ns'``, ``'us'``, ``'ms'``, ``'sec'``.
            When *unit* is ``'step'``,
            the timestep is determined by the simulator (see :make:var:`COCOTB_HDL_TIMEPRECISION`).

            .. versionchanged:: 2.0
                Renamed from ``units``.

        impl:
            One of ``'auto'``, ``'gpi'``, ``'py'``.
            Specify whether the clock is implemented with a :class:`~cocotb.simulator.GpiClock` (faster), or with a Python coroutine.
            When ``'auto'`` is used (default), the fastest implementation that supports your environment and use case is picked.

            .. versionadded:: 2.0

        set_action:
            One of :class:`.Immediate`, :class:`.Deposit`, or :class:`.Force`.
            Specify the action to use when setting the clock signal value.
            Defaults to the value of :attr:`default_set_action`.

            .. versionadded:: 2.0

        period_high:
            The period of time when the clock is driven to ``1``.
            Defaults to half of the *period*.
            Must be a multiple of the time precision of the simulator and less than *period*.

            .. versionadded:: 2.0

    When *impl* is ``'auto'``, if :envvar:`COCOTB_TRUST_INERTIAL_WRITES` is defined,
    the :class:`~cocotb.simulator.GpiClock` implementation will be used.
    Otherwise, the Python coroutine implementation will be used.
    See the environment variable's documentation for more information on the consequences
    of using the simulator's inertial write mechanism.

    If you need more features like a phase shift and an asymmetric duty cycle,
    it is simple to create your own clock generator (that you then :func:`cocotb.start_soon`):

    .. code-block:: python

        async def custom_clock():
            # pre-construct triggers for performance
            high_time = Timer(high_delay, unit="ns")
            low_time = Timer(low_delay, unit="ns")
            await Timer(initial_delay, unit="ns")
            while True:
                dut.clk.value = 1
                await high_time
                dut.clk.value = 0
                await low_time

    If you also want to change the timing during simulation,
    use this slightly more inefficient example instead where
    the :class:`Timer`\ s inside the while loop are created with
    current delay values:

    .. code-block:: python

        async def custom_clock():
            while True:
                dut.clk.value = 1
                await Timer(high_delay, unit="ns")
                dut.clk.value = 0
                await Timer(low_delay, unit="ns")


        high_delay = low_delay = 100
        cocotb.start_soon(custom_clock())
        await Timer(1000, unit="ns")
        high_delay = low_delay = 10  # change the clock speed
        await Timer(1000, unit="ns")

    .. versionadded:: 1.5
        Support ``'step'`` as the *unit* argument to mean "simulator time step".

    .. versionremoved:: 2.0
        Passing ``None`` as the *unit* argument was removed, use ``'step'`` instead.

    .. versionchanged:: 2.0
        :meth:`start` now automatically calls :func:`cocotb.start_soon` and stores the Task
        on the Clock object, so that it may later be :meth:`stop`\ ped.
    """

    _impl: Impl

    default_set_action: ClassVar[Union[Type[Immediate], Type[Deposit], Type[Force]]] = (
        Deposit
    )
    """The default action used to set the clock signal value.
    One of :class:`.Immediate`, :class:`.Deposit`, or :class:`.Force`.

    .. versionadded:: 2.0
    """

    def __init__(
        self,
        signal: LogicObject,
        period: Union[float, Fraction, Decimal],
        unit: TimeUnit = "step",
        impl: Union[Impl, None] = None,
        *,
        units: None = None,
        set_action: Union[Type[Immediate], Type[Deposit], Type[Force], None] = None,
        period_high: Union[float, Fraction, Decimal, None] = None,
    ) -> None:
        self._signal = signal

        if units is not None:
            warnings.warn(
                "The 'units' argument has been renamed to 'unit'.",
                DeprecationWarning,
                stacklevel=2,
            )
            unit = units
        self._unit: TimeUnit = unit

        self._period = period
        try:
            self._period_steps = get_sim_steps(self._period, self._unit)
        except ValueError as e:
            raise ValueError(f"Bad `period`: {e}") from None

        if set_action is None:
            set_action = type(self).default_set_action
        if set_action not in (Immediate, Deposit, Force):
            raise TypeError(
                "Invalid value for `set_action`. `set_action` must be one of Immediate, Deposit, or Force"
            )
        self._set_action = set_action

        if impl is None:
            self._impl = "gpi" if _trust_inertial else "py"
        elif impl in _valid_impls:
            self._impl = impl
        else:
            valid_impls_str = ", ".join([repr(i) for i in _valid_impls])
            raise ValueError(
                f"Invalid clock impl {impl!r}, must be one of: {valid_impls_str}"
            )

        self._period_high: Union[float, Fraction, Decimal]
        if period_high is not None:
            if period_high >= self._period:
                raise ValueError("`period_high` must be strictly less than `period`.")
            self._period_high = period_high
            try:
                self._period_high_steps = get_sim_steps(self._period_high, self._unit)
            except ValueError as e:
                raise ValueError(f"Bad `period_high`: {e}") from None
        else:
            if self._period_steps % 2 != 0:
                raise ValueError(
                    "Bad `period`: Must be divisible by 2 if `period_high` is not given."
                )
            self._period_high = period / 2
            self._period_high_steps = self._period_steps // 2

        self._task: Union[Task[None], None] = None

    @property
    def signal(self) -> LogicObject:
        """The clock signal being driven."""
        return self._signal

    @property
    def period(self) -> Union[float, Fraction, Decimal]:
        """The clock period.

        The unit is :attr:`unit`.
        """
        return self._period

    @property
    def period_high(self) -> Union[float, Fraction, Decimal]:
        """The period of time when the clock is driven to ``1``.

        The unit is :attr:`unit`.

        .. versionadded:: 2.0
        """
        return self._period_high

    @property
    def unit(self) -> TimeUnit:
        """The unit of the clock period.

        .. versionadded:: 2.0
        """
        return self._unit

    @property
    def impl(self) -> Impl:
        """The concrete implementation of the clock used.

        ``"gpi"`` if the clock is implemented in C in the GPI layer,
        or ``"py"`` if the clock is implemented in Python using cocotb Tasks.

        .. versionadded:: 2.0
        """
        return self._impl

    @property
    def set_action(self) -> Union[Type[Immediate], Type[Deposit], Type[Force]]:
        """The value setting action used to set the clock signal value.

        .. versionadded:: 2.0
        """
        return self._set_action

    def start(self, start_high: bool = True) -> Task[None]:
        r"""Start driving the clock signal.

        You can later stop the clock by calling :meth:`stop`.

        Args:
            start_high: Whether to start the clock with a ``1``
                for the first half of the period.
                Default is ``True``.

                .. versionadded:: 1.3

        Raises:
            RuntimeError: If attempting to start a clock that has already been started.

        Returns:
            Object which can be passed to :func:`cocotb.start_soon` or ignored.

        .. versionremoved:: 2.0
            Removed ``cycles`` arguments for toggling for a finite amount of cycles.
            Use :meth:`stop` to stop a clock from running.

        .. versionchanged:: 2.0
            Previously, this method returned a :term:`coroutine` which needed to be passed to :func:`cocotb.start_soon`.
            Now the Clock object keeps track of its own driver Task, so this is no longer necessary.
            Simply call ``clock.start()`` to start running the clock.
        """
        if self._task is not None:
            raise RuntimeError("Starting clock that has already been started.")

        if self._impl == "gpi":
            clkobj = clock_create(self._signal._handle)
            set_action = {
                Deposit: _GPISetAction.DEPOSIT,
                Immediate: _GPISetAction.NO_DELAY,
                Force: _GPISetAction.FORCE,
            }[self._set_action]
            clkobj.start(
                self._period_steps,
                self._period_high_steps,
                start_high,
                set_action.value,
            )

            async def drive() -> None:
                # The clock is meant to toggle forever, so awaiting this should
                # never return by awaiting on Event that's never set.
                e = Event()
                try:
                    await e.wait()
                finally:
                    clkobj.stop()

        else:

            async def drive() -> None:
                timer_high = Timer(self._period_high_steps)
                timer_low = Timer(self._period_steps - self._period_high_steps)
                if start_high:
                    self._signal.set(self._set_action(1))
                    await timer_high
                while True:
                    self._signal.set(self._set_action(0))
                    await timer_low
                    self._signal.set(self._set_action(1))
                    await timer_high

        self._task = cocotb.start_soon(drive())
        return self._task

    def stop(self) -> None:
        """Stop driving the clock signal.

        You can later start the clock again by calling :meth:`start`.

        Raises:
            RuntimeError: If attempting to stop a clock that has never been started.

        .. versionadded:: 2.0
        """
        if self._task is None:
            raise RuntimeError("Stopping a clock that was never started.")
        self._task.cancel()
        self._task = None

    async def cycles(
        self,
        num_cycles: int,
        edge_type: Union[
            Type[RisingEdge], Type[FallingEdge], Type[ValueChange]
        ] = RisingEdge,
    ) -> None:
        """Wait for a number of clock cycles.

        Args:
            num_cycles: The number of clock cycles to wait.
            edge_type:
                The edge of the clock to wait on.
                Must be one of :class:`.RisingEdge`, :class:`.FallingEdge`, or :class:`.ValueChange`.

        Raises:
            ValueError: if *num_cycles* is negative.
        """
        if num_cycles == 0:
            await NullTrigger()
            return
        elif num_cycles < 0:
            raise ValueError("`num_cycles` cannot be negative")

        # Synchronize first.
        await edge_type(self._signal)
        num_cycles -= 1

        # Use Timer as an optimization if we are waiting long enough.
        if num_cycles >= 2:
            # NOTE: num_cycles must end 1 higher than expected because all edge_types occur
            # strictly after beginning of time steps, so the last edge_type will jump within
            # the same time step.
            if edge_type is ValueChange:
                # Make cycles_skipped the nearest even number so division by 2 doesn't cause issues.
                cycles_skipped = (num_cycles // 2) * 2
                await Timer(self._period_steps * cycles_skipped / 2, "step")
                num_cycles = num_cycles - cycles_skipped + 1
            else:
                await Timer(self._period_steps * num_cycles, "step")
                num_cycles = 1

        # Run N edge_type trigger.
        for _ in range(num_cycles):
            await edge_type(self._signal)

    def __repr__(self) -> str:
        return self._repr

    @cached_property
    def _repr(self) -> str:
        freq_mhz = 1 / get_time_from_sim_steps(
            get_sim_steps(self._period, self._unit), "us"
        )
        return f"<{type(self).__qualname__}, {self._signal._path} @ {freq_mhz} MHz>"

    @cached_property
    def _log(self) -> Logger:
        return logging.getLogger(
            f"cocotb.{type(self).__qualname__}.{self._signal._name}"
        )



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
from typing import Any, Dict, TypeVar

import IPython
from IPython.terminal.ipapp import load_default_config
from IPython.terminal.prompts import Prompts
from pygments.token import Token

import cocotb
from cocotb.task import bridge
from cocotb.utils import get_sim_time

T = TypeVar("T")


class SimTimePrompt(Prompts):
    """custom prompt that shows the sim time after a trigger fires"""

    _show_time = 1

    def in_prompt_tokens(self):
        tokens = super().in_prompt_tokens()
        if self._show_time == self.shell.execution_count:
            tokens = [
                (Token.Comment, f"sim time: {get_sim_time()}"),
                (Token.Text, "\n"),
                *tokens,
            ]
        return tokens


async def embed(user_ns: Dict[str, Any] = {}) -> None:
    """
    Start an IPython shell in the current coroutine.

    Unlike using :func:`IPython.embed` directly, the :keyword:`await` keyword
    can be used directly from the shell to wait for triggers.
    The :keyword:`yield` keyword from the legacy :ref:`yield-syntax` is not supported.

    This coroutine will complete only when the user exits the interactive session.

    Args:
        user_ns:
            The variables to have made available in the shell.
            Passing ``locals()`` is often a good idea.
            ``cocotb`` will automatically be included.

    .. note::
        If your simulator does not provide an appropriate ``stdin``, you may
        find you cannot type in the resulting shell. Using simulators in batch
        or non-GUI mode may resolve this. This feature is experimental, and
        not all simulators are supported.
    """
    # ensure cocotb is in the namespace, for convenience
    default_ns = {"cocotb": cocotb}
    default_ns.update(user_ns)

    def _runner(x):
        """Handler for async functions"""
        nonlocal shell
        ret = cocotb._scheduler_inst._queue_function(x)
        shell.prompts._show_time = shell.execution_count
        return ret

    # build the config to enable `await`
    c = load_default_config()
    c.TerminalInteractiveShell.loop_runner = _runner
    c.TerminalInteractiveShell.autoawait = True
    # Python3 checks SQLite DB accesses to ensure process ID matches the one that opened the DB and is not propagated
    # because we launch IPython in a different process, this will cause unnecessary warnings, so disable the PID check
    c.HistoryAccessor.connection_options = {"check_same_thread": False}
    # create a shell with access to the dut, and cocotb pre-imported
    shell = IPython.terminal.embed.InteractiveShellEmbed(
        user_ns=default_ns,
        config=c,
    )

    # add our custom prompts
    shell.prompts = SimTimePrompt(shell)

    # start the shell in a background thread
    @bridge
    def run_shell() -> None:
        shell()

    await run_shell()


@cocotb.test()
async def run_ipython(dut: Any) -> None:
    """A test that launches an interactive Python shell.

    Do not call this directly - use this as ``make COCOTB_TEST_MODULES=cocotb.ipython_support``.

    Within the shell, a global ``dut`` variable pointing to the design will be present.
    """
    await embed(user_ns={"dut": dut})



# Copyright cocotb contributors
# Copyright (c) 2013, 2018 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""Task scheduler."""

import logging
import threading
from bdb import BdbQuit
from collections import OrderedDict
from typing import Any, Callable, Coroutine, Dict, List, TypeVar, Union

import cocotb
import cocotb._gpi_triggers
import cocotb.handle
from cocotb import debug
from cocotb._base_triggers import Event, Trigger
from cocotb._bridge import external_state, external_waiter
from cocotb._exceptions import InternalError
from cocotb._gpi_triggers import (
    GPITrigger,
    NextTimeStep,
    ReadWrite,
)
from cocotb._outcomes import Error, Outcome, Value, capture
from cocotb._profiling import profiling_context
from cocotb._py_compat import ParamSpec, insertion_ordered_dict
from cocotb.task import Task, _TaskState

T = TypeVar("T")

P = ParamSpec("P")


class Scheduler:
    """The main Task scheduler.

    How It Generally Works:
        Tasks are `queued` to run in the scheduler with :meth:`_queue`.
        Queueing adds the Task and an Outcome value to :attr:`_pending_tasks`.
        The main scheduling loop is located in :meth:`_event_loop` and loops over the queued Tasks and `schedules` them.
        :meth:`_schedule` schedules a Task -
        continuing its execution from where it previously yielded control -
        by injecting the Outcome value associated with the Task from the queue.
        The Task's body will run until it finishes or reaches the next :keyword:`await` statement.
        If a Task reaches an :keyword:`await`, :meth:`_schedule` will convert the value yielded from the Task into a Trigger with :meth:`_trigger_from_any` and its friend methods.
        Triggers are then `primed` (with :meth:`~cocotb.triggers.Trigger._prime`)
        with a `react` function (:meth:`_sim_react` or :meth:`_react)
        so as to wake up Tasks waiting for that Trigger to `fire` (when the event encoded by the Trigger occurs).
        This is accomplished by :meth:`_resume_task_upon`.
        :meth:`_resume_task_upon` also associates the Trigger with the Task waiting on it to fire by adding them to the :attr:`_trigger2tasks` map.
        If, instead of reaching an :keyword:`await`, a Task finishes, :meth:`_schedule` will cause the :class:`~cocotb.task.Join` trigger to fire.
        Once a Trigger fires it calls the react function which queues all Tasks waiting for that Trigger to fire.
        Then the process repeats.

        When a Task is cancelled (:meth:`_unschedule`), it is removed from the Task queue if it is currently queued.
        Also, the Task and Trigger are deassociated in the :attr:`_trigger2tasks` map.
        If the cancelled Task is the last Task waiting on a Trigger, that Trigger is `unprimed` to prevent it from firing.

    Simulator Phases:
        All GPITriggers (triggers that are fired by the simulator) go through :meth:`_sim_react`
        which looks at the fired GPITriggers to determine and track the current simulator phase cocotb is executing in.

        Normal phase:
            Corresponds to all non-ReadWrite and non-ReadOnly phases.
            Any writes are cached for the next ReadWrite phase and do not happen immediately.
            Scheduling :class:`~cocotb.triggers.ReadWrite` and :class:`~cocotb.triggers.ReadOnly` are valid.

        ReadWrite phase:
            Corresponds to ``cbReadWriteSynch`` (VPI) or ``vhpiCbRepLastKnownDeltaCycle`` (VHPI).
            At the start of scheduling in this phase we play back all the *previously* cached write updates.
            Any writes are cached for the next ReadWrite phase and do not happen immediately.
            Scheduling :class:`~cocotb.triggers.ReadWrite` and :class:`~cocotb.triggers.ReadOnly` are valid.
            One caveat is that scheduling a :class:`~cocotb.triggers.ReadWrite` while in this phase may not be valid.
            If there were no writes applied at the beginning of this phase, there will be no more events in this time step,
            and there will not be another ReadWrite phase in this time step.
            Simulators generally handle this caveat gracefully by leaving you in the ReadWrite phase of the next time step.

        ReadOnly phase
            Corresponds to ``cbReadOnlySynch`` (VPI) or ``vhpiCbRepEndOfTimeStep`` (VHPI).
            In this state we are not allowed to perform writes.
            Scheduling :class:`~cocotb.triggers.ReadWrite` and :class:`~cocotb.triggers.ReadOnly` are *not* valid.

    Caveats and Special Cases:
        The scheduler treats Tests specially.
        If a Test finishes or a Task ends with an Exception, the scheduler is put into a `terminating` state.
        All currently queued Tasks are cancelled and all pending Triggers are unprimed.
        This is currently spread out between :meth:`_handle_termination` and :meth:`_cleanup`.
        In that mix of functions, the :attr:`_test_complete_cb` callback is called to inform whomever (the regression_manager) the test finished.
        The scheduler also is responsible for starting the next Test in the Normal phase by priming a ``Timer(1)`` with the second half of test completion handling.

        The scheduler is currently where simulator time phase is tracked.
        This is mostly because this is where :meth:`_sim_react` is most conveniently located.
        The scheduler can't currently be made independent of simulator-specific code because of the above special cases which have to respect simulator phasing.

        Currently Task cancellation is accomplished with :meth:`Task.kill() <cocotb.task.Task.kill>`.
        This function immediately cancels the Task by re-entering the scheduler.
        This can cause issues if you are trying to cancel the Test Task or the currently executing Task.

        TODO: There are attributes and methods for dealing with "externals", but I'm not quite sure how it all works yet.
    """

    # Singleton events, recycled to avoid spurious object creation
    _next_time_step = NextTimeStep()
    _read_write = ReadWrite()

    def __init__(self) -> None:
        self.log = logging.getLogger("cocotb.scheduler")

        # A dictionary of pending tasks for each trigger,
        # indexed by trigger
        self._trigger2tasks: Dict[Trigger, list[Task[object]]] = (
            insertion_ordered_dict()
        )

        self._scheduled_tasks: OrderedDict[Task[object], Union[BaseException, None]] = (
            OrderedDict()
        )
        self._pending_threads: List[external_waiter[Any]] = []
        self._pending_events: List[Event] = []

        self._main_thread = threading.current_thread()

        self._current_task: Union[Task[object], None] = None

    def _sim_react(self, trigger: GPITrigger) -> None:
        """Called when a :class:`~cocotb.triggers.GPITrigger` fires.

        This is often the entry point into Python from the simulator,
        so this function is in charge of enabling profiling.
        It must also track the current simulator time phase,
        and start the unstarted event loop.
        """
        with profiling_context:
            # TODO: move state tracking to global variable
            # and handle this via some kind of trigger-specific Python callback
            cocotb._gpi_triggers._current_gpi_trigger = trigger

            # apply inertial writes if ReadWrite
            if trigger is self._read_write:
                cocotb.handle._apply_scheduled_writes()

            self._react(trigger)
            self._event_loop()

    def _react(self, trigger: Trigger) -> None:
        """Called when a :class:`~cocotb.triggers.Trigger` fires.

        Finds all Tasks waiting on the Trigger that fired and queues them.
        """
        if debug.debug:
            self.log.debug("Trigger fired: %s", trigger)

        # find all tasks waiting on trigger that fired
        try:
            scheduling = self._trigger2tasks.pop(trigger)
        except KeyError:
            # GPI triggers should only be ever pending if there is an
            # associated task waiting on that trigger, otherwise it would
            # have been unprimed already
            if isinstance(trigger, GPITrigger):
                self.log.warning(
                    "No tasks waiting on GPITrigger that fired: %s\n"
                    "This is due to an issue with the GPI or a simulator bug.",
                    trigger,
                )
            # For Python triggers this isn't actually an error - we might do
            # event.set() without knowing whether any tasks are actually
            # waiting on this event, for example
            elif debug.debug:
                self.log.debug("No tasks waiting on trigger that fired: %s", trigger)
            return

        if debug.debug:
            debugstr = "\n\t".join([str(task) for task in scheduling])
            if len(scheduling) > 0:
                debugstr = "\n\t" + debugstr
            self.log.debug(
                "%d pending tasks for trigger %s%s",
                len(scheduling),
                trigger,
                debugstr,
            )

        # queue all tasks to wake up
        for task in scheduling:
            # unset trigger
            task._trigger = None
            self._schedule_task_internal(task)

        # cleanup trigger
        trigger._cleanup()

    def _event_loop(self) -> None:
        """Run the main event loop.

        This should only be started by:
        * The beginning of a test, when there is no trigger to react to
        * A GPI trigger
        """

        while self._scheduled_tasks:
            task, exc = self._scheduled_tasks.popitem(last=False)

            if debug.debug:
                self.log.debug("Scheduling task %s", task)
            self._resume_task(task, exc)
            if debug.debug:
                self.log.debug("Scheduled task %s", task)

            # remove our reference to the objects at the end of each loop,
            # to try and avoid them being destroyed at a weird time (as
            # happened in gh-957)
            del task

            # Schedule may have queued up some events so we'll burn through those
            while self._pending_events:
                if debug.debug:
                    self.log.debug(
                        "Scheduling pending event %s", self._pending_events[0]
                    )
                self._pending_events.pop(0).set()

        # no more pending tasks
        if debug.debug:
            self.log.debug("All tasks scheduled, handing control back to simulator")

    def _unschedule(self, task: Task[Any]) -> None:
        """Unschedule a task and unprime dangling pending triggers.

        Also:
          * enters the scheduler termination state if the Test Task is unscheduled.
          * creates and fires a :class:`~cocotb.task.Join` trigger.
          * forcefully ends the Test if a Task ends with an exception.
        """

        # remove task from queue
        if task in self._scheduled_tasks:
            self._scheduled_tasks.pop(task)

        # Unprime the trigger this task is waiting on
        trigger = task._trigger
        if trigger is not None:
            task._trigger = None
            if task in self._trigger2tasks.setdefault(trigger, []):
                self._trigger2tasks[trigger].remove(task)
            if not self._trigger2tasks[trigger]:
                trigger._unprime()
                del self._trigger2tasks[trigger]

    def _schedule_task_upon(self, task: Task[Any], trigger: Trigger) -> None:
        """Schedule `task` to be resumed when `trigger` fires."""
        # TODO Move this all into Task
        task._trigger = trigger
        task._state = _TaskState.PENDING

        trigger_tasks = self._trigger2tasks.setdefault(trigger, [])
        trigger_tasks.append(task)

        try:
            # TODO maybe associate the react method with the trigger object so
            # we don't have to do a type check here.
            if isinstance(trigger, GPITrigger):
                trigger._prime(self._sim_react)
            else:
                trigger._prime(self._react)
        except Exception as e:
            # discard the trigger we associated, it will never fire
            self._trigger2tasks.pop(trigger)

            # replace it with a new trigger that throws back the exception
            self._schedule_task_internal(task, e)

    def _schedule_task(self, task: Task[Any]) -> None:
        """Queue *task* for scheduling.

        It is an error to attempt to queue a task that has already been queued.
        """
        if task.done():
            raise RuntimeError(
                f"{task} has finished executing and can not be scheduled again. Did you call start_soon() on a finished Task?"
            )
        if task in self._scheduled_tasks:
            return
        for tasks in self._trigger2tasks.values():
            if task in tasks:
                return
        self._schedule_task_internal(task)

    def _schedule_task_internal(
        self, task: Task[Any], exc: Union[BaseException, None] = None
    ) -> None:
        # TODO Move state tracking into Task
        task._state = _TaskState.SCHEDULED
        self._scheduled_tasks[task] = exc

    def _queue_function(self, task: Coroutine[Trigger, None, T]) -> T:
        """Queue a task for execution and move the containing thread
        so that it does not block execution of the main thread any longer.
        """
        # We should be able to find ourselves inside the _pending_threads list
        matching_threads = [
            t for t in self._pending_threads if t.thread == threading.current_thread()
        ]
        if len(matching_threads) == 0:
            raise RuntimeError("queue_function called from unrecognized thread")

        # Raises if there is more than one match. This can never happen, since
        # each entry always has a unique thread.
        (t,) = matching_threads

        outcome: Union[Outcome[T], None] = None

        async def wrapper() -> None:
            nonlocal outcome
            # This function runs in the scheduler thread
            try:
                outcome = Value(await task)
            except (KeyboardInterrupt, SystemExit, BdbQuit):
                # Allow these to bubble up to the execution root to fail the sim immediately.
                # This follows asyncio's behavior.
                raise
            except BaseException as e:
                outcome = Error(e)
            # Notify the current (scheduler) thread that we are about to wake
            # up the background (`@external`) thread, making sure to do so
            # before the background thread gets a chance to go back to sleep by
            # calling thread_suspend.
            # We need to do this here in the scheduler thread so that no more
            # tasks run until the background thread goes back to sleep.
            t.thread_resume()
            event.set()

        event = threading.Event()
        # must register this with test as there's no way to clean up with threading
        self._schedule_task_internal(cocotb.start_soon(wrapper()))
        # The scheduler thread blocks in `thread_wait`, and is woken when we
        # call `thread_suspend` - so we need to make sure the task is
        # queued before that.
        t.thread_suspend()
        # This blocks the calling `@external` thread until the task finishes
        event.wait()
        assert outcome is not None
        return outcome.get()

    def _run_in_executor(
        self, func: "Callable[P, T]", *args: "P.args", **kwargs: "P.kwargs"
    ) -> Coroutine[Trigger, None, T]:
        """Run the task in a separate execution thread
        and return an awaitable object for the caller.
        """
        # Create a thread
        # Create a trigger that is called as a result of the thread finishing
        # Create an Event object that the caller can await on
        # Event object set when the thread finishes execution, this blocks the
        # calling task (but not the thread) until the external completes

        waiter = external_waiter[T]()

        def execute_external() -> None:
            waiter._outcome = capture(func, *args, **kwargs)
            if debug.debug:
                self.log.debug(
                    "Execution of external routine done %s", threading.current_thread()
                )
            waiter.thread_done()

        async def wrapper() -> T:
            thread = threading.Thread(
                group=None,
                target=execute_external,
                name=func.__qualname__ + "_thread",
            )

            waiter.thread = thread
            self._pending_threads.append(waiter)

            await waiter.event.wait()

            return waiter.result  # raises if there was an exception

        return wrapper()

    def _resume_task(self, task: Task[object], exc: Union[BaseException, None]) -> None:
        """Resume *task* with *outcome*.

        Args:
            task: The task to schedule.
            outcome: The outcome to inject into the *task*.

        Scheduling runs *task* until it either finishes or reaches the next :keyword:`await` statement.
        If *task* completes, it is unscheduled, a Join trigger fires, and test completion is inspected.
        Otherwise, it reached an :keyword:`await` and we have a result object which is converted to a trigger,
        that trigger is primed,
        then that trigger and the *task* are registered with the :attr:`_trigger2tasks` map.
        """
        if self._current_task is not None:
            raise InternalError("_schedule() called while another Task is executing")
        try:
            self._current_task = task

            trigger = task._advance(exc)

            if task.done():
                if debug.debug:
                    self.log.debug("%s completed with %s", task, task._outcome)
                assert trigger is None
                self._unschedule(task)

            if not task.done():
                if debug.debug:
                    self.log.debug("%r yielded %s", task, trigger)
                if not isinstance(trigger, Trigger):
                    e = TypeError(
                        f"Coroutine yielded an object of type {type(trigger)}, which the scheduler can't "
                        f"handle: {trigger!r}\n"
                    )
                    self._schedule_task_internal(task, e)
                else:
                    self._schedule_task_upon(task, trigger)

            # We do not return from here until pending threads have completed, but only
            # from the main thread, this seems like it could be problematic in cases
            # where a sim might change what this thread is.

            if self._main_thread is threading.current_thread():
                for ext in self._pending_threads:
                    ext.thread_start()
                    if debug.debug:
                        self.log.debug(
                            "Blocking from %s on %s",
                            threading.current_thread(),
                            ext.thread,
                        )
                    state = ext.thread_wait()
                    if debug.debug:
                        self.log.debug(
                            "Back from wait on self %s with newstate %s",
                            threading.current_thread(),
                            state,
                        )
                    if state == external_state.EXITED:
                        self._pending_threads.remove(ext)
                        self._pending_events.append(ext.event)
        finally:
            self._current_task = None



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

import enum
import logging
import os
import re
from abc import ABC, abstractmethod
from functools import lru_cache
from logging import Logger
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import cocotb
from cocotb import simulator
from cocotb._base_triggers import Event
from cocotb._deprecation import deprecated
from cocotb._gpi_triggers import (
    Edge,
    FallingEdge,
    ReadOnly,
    ReadWrite,
    RisingEdge,
    ValueChange,
    current_gpi_trigger,
)
from cocotb._py_compat import cached_property, insertion_ordered_dict
from cocotb._utils import DocIntEnum
from cocotb.task import Task
from cocotb.types import Array, Logic, LogicArray, Range
from cocotb.types._indexing import do_indexing_changed_warning, indexing_changed

__all__ = (
    "ArrayObject",
    "Deposit",
    "EnumObject",
    "Force",
    "Freeze",
    "GPIDiscovery",
    "HierarchyArrayObject",
    "HierarchyObject",
    "Immediate",
    "IntegerObject",
    "LogicArrayObject",
    "LogicObject",
    "RealObject",
    "Release",
    "SimHandleBase",
    "StringObject",
    "ValueObjectBase",
)


class _Limits(enum.IntEnum):
    SIGNED_NBIT = 1
    UNSIGNED_NBIT = 2
    VECTOR_NBIT = 3


@lru_cache(maxsize=None)
def _value_limits(n_bits: int, limits: _Limits) -> Tuple[int, int]:
    """Calculate min/max for given number of bits and limits class"""
    if limits == _Limits.SIGNED_NBIT:
        min_val = -(2 ** (n_bits - 1))
        max_val = 2 ** (n_bits - 1) - 1
    elif limits == _Limits.UNSIGNED_NBIT:
        min_val = 0
        max_val = 2**n_bits - 1
    else:
        min_val = -(2 ** (n_bits - 1))
        max_val = 2**n_bits - 1

    return min_val, max_val


class SimHandleBase(ABC):
    """Base class for all simulation objects.

    All simulation objects are hashable and equatable by identity.

    .. code-block:: python

        a = dut.clk
        b = dut.clk
        assert a == b

    .. versionchanged:: 2.0
        ``get_definition_name()`` and ``get_definition_file()`` were removed in favor of :meth:`_def_name` and :meth:`_def_file`, respectively.
    """

    @abstractmethod
    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        self._handle = handle
        self._path: str = self._name if path is None else path
        """The path to this handle, or its name if this is the root handle.

        :meta public:
        """

    @cached_property
    def _name(self) -> str:
        """The name of an object.

        :meta public:
        """
        return self._handle.get_name_string()

    @cached_property
    def _type(self) -> str:
        """The type of an object as a string.

        :meta public:
        """
        return self._handle.get_type_string()

    @cached_property
    def _log(self) -> Logger:
        return logging.getLogger(f"cocotb.{self._name}")

    @cached_property
    def _def_name(self) -> str:
        """The name of a GPI object's definition.

        This is the value of ``vpiDefName`` for VPI, ``vhpiNameP`` for VHPI,
        and ``mti_GetPrimaryName`` for FLI.
        Support for this depends on the specific object type and simulator used.

        :meta public:
        """
        return self._handle.get_definition_name()

    @cached_property
    def _def_file(self) -> str:
        """The name of the file that sources the object's definition.

        This is the value of ``vpiDefFile`` for VPI, ``vhpiFileNameP`` for VHPI,
        and ``mti_GetRegionSourceName`` for FLI.
        Support for this depends on the specific object type and simulator used.

        :meta public:
        """
        return self._handle.get_definition_file()

    def __hash__(self) -> int:
        return hash(self._handle)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SimHandleBase):
            return NotImplemented
        return self._handle == other._handle

    def __repr__(self) -> str:
        desc = self._path
        defname = self._def_name
        if defname:
            desc += " with definition " + defname
            deffile = self._def_file
            if deffile:
                desc += " (at " + deffile + ")"
        return type(self).__qualname__ + "(" + desc + ")"

    def __bool__(self) -> NoReturn:
        raise TypeError(
            "This object cannot be cast to bool or used in conditionals. Use `obj is not None` check in conditionals."
        )


class _RangeableObjectMixin(SimHandleBase):
    """Base class for simulation objects that have a range."""

    @cached_property
    def range(self) -> Range:
        """Return a :class:`~cocotb.types.Range` over the indexes of the array/vector."""
        left, right, direction = self._handle.get_range()
        if direction == simulator.RANGE_NO_DIR:
            raise RuntimeError("Expected range to have a direction but got none!")
        return Range(left, "to" if direction == simulator.RANGE_UP else "downto", right)

    @property
    def left(self) -> int:
        """Return the leftmost index in the array/vector."""
        return self.range.left

    @property
    def direction(self) -> str:
        """Return the direction (``"to"``/``"downto"``) of indexes in the array/vector."""
        return self.range.direction

    @property
    def right(self) -> int:
        """Return the rightmost index in the array/vector."""
        return self.range.right

    def __len__(self) -> int:
        return len(self.range)


#: Type of keys (name or index) in HierarchyObjectBase.
KeyType = TypeVar("KeyType")

#: Subtype of :class:`SimHandleBase` returned when iterating or indexing a :class:`HierarchyArrayObject`.
HierarchyChildObjectT = TypeVar("HierarchyChildObjectT", bound=SimHandleBase)


class GPIDiscovery(DocIntEnum):
    """Simulator object discovery strategy."""

    AUTO = (0, "Automatic discovery using all registered interfaces.")
    NATIVE = (1, "Native discovery using only the parent's native interface.")


class _HierarchyObjectBase(SimHandleBase, Generic[KeyType]):
    """Base class for hierarchical simulation objects.

    Hierarchical objects don't have values, they are just scopes/namespaces of other objects.
    This includes array-like hierarchical structures like "generate loops"
    and named hierarchical structures like "generate blocks" or "module"/"entity" instantiations.

    This base class defines logic to discover, cache, and inspect child objects.
    It provides a :class:`dict`-like interface for doing so.

    :meth:`_keys`, :meth:`_values`, and :meth:`_items` mimic their :class:`dict` counterparts.
    You can also iterate over an object, which returns child objects, not keys like in :class:`dict`;
    and can check the :func:`len`.

    See :class:`HierarchyObject` and :class:`HierarchyArrayObject` for examples.
    """

    @abstractmethod
    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        super().__init__(handle, path)
        self._sub_handles: Dict[KeyType, SimHandleBase] = {}
        self._discovered = False

    def _keys(self) -> Iterable[KeyType]:
        """Iterate over the keys (name or index) of the child objects.

        :meta public:
        """
        self._discover_all()
        return self._sub_handles.keys()

    def _values(self) -> Iterable[SimHandleBase]:
        """Iterate over the child objects.

        :meta public:
        """
        self._discover_all()
        return self._sub_handles.values()

    def _items(self) -> Iterable[Tuple[KeyType, SimHandleBase]]:
        """Iterate over ``(key, object)`` tuples of child objects.

        :meta public:
        """
        self._discover_all()
        return self._sub_handles.items()

    def _discover_all(self) -> None:
        """When iterating or performing IPython tab completion, we run through ahead of
        time and discover all possible children, populating the :any:`_sub_handles`
        mapping. Hierarchy can't change after elaboration so we only have to
        do this once.
        """
        if self._discovered:
            return

        for thing in self._handle.iterate(simulator.OBJECTS):
            name = thing.get_name_string()

            # translate HDL name into a consistent key name
            try:
                key = self._sub_handle_key(name)
            except ValueError:
                self._log.exception(
                    "Unable to translate handle >%s< to a valid _sub_handle key",
                    name,
                )
                continue

            # compute a full path using the key name
            path = self._child_path(key)

            # attempt to create the child object
            try:
                hdl = _make_sim_object(thing, path)
            except NotImplementedError:
                self._log.exception(
                    "Unable to construct a SimHandle object for %s", path
                )
                continue

            # add to cache
            self._sub_handles[key] = hdl

        self._discovered = True

    def _get(
        self, key: KeyType, discovery_method: GPIDiscovery = GPIDiscovery.AUTO
    ) -> Union[SimHandleBase, None]:
        """Query the simulator for an object with the specified *key*.

        Like Python's native dictionary ``get``-function, this returns ``None`` if the object
        is not found instead of raising an :exc:`AttributeError`.

        Generally, use the ``handle[child_name]`` syntax instead, unless you have to change the
        *discovery_method* or want to check for optional signals.

        :meta public:

        Args:
            key: The child object by name.
            discovery_method: Optional selection of discovery strategy. :data:`~cocotb.handle.GPIDiscovery.AUTO` by default.

        Returns:
            The child object, or ``None`` if not found.
        """
        # try to use cached value
        try:
            return self._sub_handles[key]
        except KeyError:
            pass

        # try to get value from GPI
        new_handle = self._get_handle_by_key(key, discovery_method)
        if new_handle is None:
            return None

        # if successful, construct and cache
        sub_handle = _make_sim_object(new_handle, self._child_path(key))
        self._sub_handles[key] = sub_handle

        return sub_handle

    @abstractmethod
    def _get_handle_by_key(
        self, key: KeyType, discovery_method: GPIDiscovery
    ) -> Union[simulator.gpi_sim_hdl, None]:
        """Get child object by key from the simulator.

        Args:
            key: The key of the child object.
            discovery_method: How to discover the object using the GPI.

        Returns:
            A raw simulator handle for the child object at the given key, or ``None``.
        """

    @abstractmethod
    def _child_path(self, key: KeyType) -> str:
        """Compute the path string of a child object at the given key.

        Args:
            key: The key of the child object.

        Returns:
            A path string of the child object at the a given key.
        """

    @abstractmethod
    def _sub_handle_key(self, name: str) -> KeyType:
        """Translate a discovered child object name into a key.

        Args:
            name: The GPI name of the child object.

        Returns:
            A unique key for the child object.

        Raises:
            ValueError: if unable to translate handle to a valid _sub_handle key.
        """

    def __iter__(self) -> Iterator[SimHandleBase]:
        return iter(self._values())

    def __len__(self) -> int:
        self._discover_all()
        return len(self._sub_handles)

    def __dir__(self) -> Iterable[str]:
        """Permits IPython tab completion and debuggers to work."""
        self._discover_all()
        return set(super().__dir__()) | {str(k) for k in self._keys()}


class HierarchyObject(_HierarchyObjectBase[str]):
    r"""A simulation object that is a name-indexed collection of hierarchical simulation objects.

    Inherits from :class:`SimHandleBase`.

    This class is used for named hierarchical structures, such as "generate blocks" or "module"/"entity" instantiations.

    Children under this structure are found by using the name of the child with either the attribute syntax or index syntax.
    For example, if in your :envvar:`COCOTB_TOPLEVEL` entity/module you have a signal/net named ``count``, you could do either of the following.

    .. code-block:: python

        dut.count  # attribute syntax
        dut["count"]  # index syntax

    Attribute syntax is usually shorter and easier to read, and is more common.
    However, it has limitations:

    - the name cannot start with a number
    - the name cannot start with a ``_`` character
    - the name can only contain ASCII letters, numbers, and the ``_`` character.

    Any possible name of an object is supported with the index syntax,
    but it can be more verbose.

    Accessing a non-existent child with attribute syntax results in an :class:`AttributeError`,
    and accessing a non-existent child with index syntax results in a :class:`KeyError`.

    .. note::
        If you need to access signals/nets that start with ``_``,
        or use escaped identifier (Verilog) or extended identifier (VHDL) characters,
        you have to use the index syntax.
        Accessing escaped/extended identifiers requires enclosing the name
        with leading and trailing double backslashes (``\\``).

        .. code-block:: python

            dut["_underscore_signal"]
            dut["\\%extended !ID\\"]

    Iteration yields all child objects in no particular order.
    The :func:`len` function can be used to find the number of children.

    .. code-block:: python

        # discover all children in 'some_module'
        total = 0
        for handle in dut.some_module:
            cocotb.log("Found %r", handle._path)
            total += 1

        # make sure we found them all
        assert len(dut.some_module) == total
    """

    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        super().__init__(handle, path)

    def __setattr__(self, name: str, value: object) -> None:
        # private attributes pass through directly
        if name.startswith("_"):
            return object.__setattr__(self, name, value)

        try:
            getattr(self, name)
        except AttributeError:
            raise AttributeError(
                f"Cannot set attribute {name!r} on simulation object {self._path}. No such object exists."
            ) from None
        else:
            raise AttributeError(
                f"Cannot set attribute {name!r} on simulation object {self._path}. Did you forget to add `.value`?"
            )

    def __getattr__(self, name: str) -> SimHandleBase:
        if name.startswith("_"):
            return object.__getattribute__(self, name)

        handle = self._get(name)
        if handle is None:
            raise AttributeError(f"{self._path} contains no child object named {name}")
        return handle

    def __getitem__(self, key: str) -> SimHandleBase:
        handle = self._get(key)
        if handle is None:
            raise KeyError(f"{self._path} contains no child object named {key}")
        return handle

    @deprecated(
        "Use `handle[child_name]` syntax instead. If extended identifiers are needed simply add a '\\' character before and after the name."
    )
    def _id(self, name: str, extended: bool = True) -> SimHandleBase:
        """Query the simulator for an object with the specified *name*.

        If *extended* is ``True``, run the query only for VHDL extended identifiers.
        For Verilog, only ``extended=False`` is supported.

        :meta public:

        Args:
            name: The child object by name.
            extended: If ``True``, treat the *name* as an extended identifier.

        Returns:
            The child object.

        Raises:
            AttributeError: If the child object is not found.

        .. deprecated:: 2.0
            Use ``handle[child_name]`` syntax instead.
            If extended identifiers are needed simply add a ``\\`` character before and after the name.

        """
        if extended:
            name = "\\" + name + "\\"

        handle = self._get(name)
        if handle is None:
            raise AttributeError(f"{self._path} contains no child object named {name}")
        return handle

    def _child_path(self, key: str) -> str:
        delimiter = "::" if self._type == "GPI_PACKAGE" else "."
        return f"{self._path}{delimiter}{key}"

    def _sub_handle_key(self, name: str) -> str:
        return name

    def _get_handle_by_key(
        self, key: str, discovery_method: GPIDiscovery
    ) -> Union[simulator.gpi_sim_hdl, None]:
        return self._handle.get_handle_by_name(key, discovery_method)


class HierarchyArrayObject(
    _HierarchyObjectBase[int], _RangeableObjectMixin, Generic[HierarchyChildObjectT]
):
    """A simulation object that is an array of hierarchical simulation objects.

    Inherits from :class:`SimHandleBase`.

    This class is used for array-like hierarchical structures like "generate loops".

    Children of this object are found by supplying a numerical index using index syntax.
    For example, if you have a design with a generate loop ``gen_pipe_stages`` from the range ``0`` to ``7``:

    .. code-block:: python

        block_0 = dut.gen_pipe_stages[0]
        block_7 = dut.gen_pipe_stages[7]

    Accessing a non-existent child results in an :class:`IndexError`.

    Iteration yields all child objects in order.

    .. code-block:: python

        # set all 'reg's in each pipe stage to 0
        for pipe_stage in dut.gen_pipe_stages:
            pipe_stage.reg.value = 0

    Use the :meth:`range` property if you want to iterate over the indexes.
    The :func:`len` function can be used to find the number of elements.

    .. code-block:: python

        # set all 'reg's in each pipe stage to 0
        for idx in dut.gen_pipe_stages.range:
            dut.gen_pipe_stages[idx].reg.value = 0

        # make sure we have all the pipe stages
        assert len(dut.gen_pipe_stage) == len(dut.gen_pipe_stages.range)
    """

    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        super().__init__(handle, path)

    def _sub_handle_key(self, name: str) -> int:
        # This is slightly hacky, but we need to extract the index from the name
        # See also GEN_IDX_SEP_* in VhpiImpl.h for the VHPI separators.
        #
        # FLI and VHPI:       _name(X) where X is the index
        # VHPI(ALDEC):        _name__X where X is the index
        # VPI:                _name[X] where X is the index
        result = re.match(rf"{re.escape(self._name)}__(?P<index>\d+)$", name)
        if not result:
            result = re.match(
                rf"{re.escape(self._name)}\((?P<index>\d+)\)$", name, re.IGNORECASE
            )
        if not result:
            result = re.match(rf"{re.escape(self._name)}\[(?P<index>\d+)\]$", name)

        if result:
            return int(result.group("index"))
        else:
            raise ValueError(f"Unable to match an index pattern: {name}")

    def _child_path(self, key: int) -> str:
        return f"{self._path}[{key}]"

    def _get_handle_by_key(
        self, key: int, discovery_method: GPIDiscovery
    ) -> Union[simulator.gpi_sim_hdl, None]:
        if discovery_method is not GPIDiscovery.AUTO:
            raise NotImplementedError(
                f"Only GPIDiscovery.AUTO is supported for {type(self).__qualname__} right now"
            )
        return self._handle.get_handle_by_index(key)

    def __getitem__(self, key: int) -> HierarchyChildObjectT:
        if isinstance(key, slice):
            raise TypeError("Slice indexing is not supported")

        handle = self._get(key)
        if handle is None:
            raise IndexError(f"{self._path} contains no child object at index {key}")
        return cast("HierarchyChildObjectT", handle)

    # ideally `__len__` could be implemented in terms of `range`, but `range` doesn't work universally.

    def __iter__(self) -> Iterator[HierarchyChildObjectT]:
        # must use `sorted(self._keys())` instead of the range because `range` doesn't work universally.
        for i in sorted(self._keys()):
            yield self[i]


class _GPISetAction(enum.Enum):
    DEPOSIT = 0
    FORCE = 1
    RELEASE = 2
    NO_DELAY = 3
    OLD_IMMEDIATE = 0


_ValueT = TypeVar("_ValueT")


class Deposit(Generic[_ValueT]):
    r""":term:`Inertially deposit <inertial deposit>` the given value on a simulator object.

    If another :term:`deposit` comes after this deposit, the newer deposit overwrites the old value.
    If an HDL process is :term:`driving` the signal/net/register where a deposit from cocotb is made,
    the deposited value will be overwritten at the end of the next delta cycle,
    essentially causing a single delta cycle "glitch" in the waveform.

    .. note::
        VHDL applies writes according to their definition.
        ``signal`` writes are set inertially, regardless of using this class;
        while ``variable`` writes are set immediately, regardless of using this class.
    """

    def __init__(self, value: _ValueT) -> None:
        self.value = value


class Force(Generic[_ValueT]):
    r""":term:`Force <force>` the given value on a simulator object immediately.

    Further :term:`deposits <deposit>` from cocotb or :term:`drives <driving>` from HDL processes
    do not cause the value to change until the handle is :term:`released <release>` by cocotb or HDL code.
    Further :term:`forces <force>` will overwrite the value and leave the value forced.

    .. note::
        VHDL applies writes according to their definition.
        ``signal`` writes are set inertially, regardless of using this class;
        while ``variable`` writes are set immediately, regardless of using this class.

    .. note::
        Verilog :class:`!Force`\ s are always immediate.
        This also means that if there are multiple cocotb Tasks or multiple ``always`` blocks writing to the same object,
        the resulting value is non-deterministic.

    .. note::
        Issuing a :class:`!Force` and :class:`Release` in the same evaluation cycle in VHDL will result in the :class:`!Force` "winning".
    """

    def __init__(self, value: _ValueT) -> None:
        self.value = value


class Freeze:
    r""":term:`Force <force>` the simulator object with its current value.

    Useful if you have done a :term:`deposit` and later decide to lock the value from changing.
    Does not change the current value of the simulation object.
    See :class:`Force` for information on behavior after this write completes.

    .. note::
        VHDL applies writes according to their definition.
        ``signal`` writes are set inertially, regardless of using this class;
        while ``variable`` writes are set immediately, regardless of using this class.

    .. note::
        Verilog :class:`Force`\ s are always immediate.
        This also means that if there are multiple cocotb Tasks or multiple ``always`` blocks writing to the same object,
        the resulting value is non-deterministic.

    .. note::
        Issuing a :class:`!Force` and :class:`Release` in the same evaluation cycle in VHDL will result in the :class:`!Force` "winning".
    """


class Release:
    r""":term:`Release <release>` a :term:`forced <force>` simulation object.

    Does not change the current value of the simulation object.
    See :class:`Deposit` for information on behavior after this write completes.

    .. note::
        VHDL applies writes according to their definition.
        ``signal`` writes are set inertially, regardless of using this class,
        while ``variable`` writes are set immediately, regardless of using this class.

    .. note::
        Verilog :class:`!Release`\ s are always immediate.
        This also means that if there are multiple cocotb Tasks or multiple ``always`` blocks writing to the same object,
        the resulting value is non-deterministic.

    .. note::
        Issuing a :class:`Force` and :class:`!Release` in the same evaluation cycle in VHDL will result in the :class:`!Force` "winning".

    .. note::
        Releasing a ``reg`` or ``logic`` in Verilog will leave the current value.
        Releasing a ``wire`` in Verilog will cause the value to be recomputed from the wire's drivers current values.
        Releasing a ``signal`` in VHDL will cause the value to be recomputed from the signal's drivers current value.
        Unconnected ``in`` ports and unconnected internal signals have no drivers and their value after :class:`!Release` will be ``U`` in VHDL and ``X`` in Verilog.
    """


class Immediate(Generic[_ValueT]):
    """:term:`Deposit <no-delay deposit>` a value on a simulator object without delay.

    The value of the signal will be changed immediately
    and should be able to be read back immediately following the write.
    Otherwise, behaves like :class:`Deposit`.

    .. note::
        VHDL applies writes according to their definition.
        ``signal`` writes are set inertially, regardless of using this class;
        while ``variable`` writes are set immediately, regardless of using this class.

    .. note::
        In Verilog, because these writes are immediate,
        if there are multiple cocotb Tasks or multiple ``always`` blocks writing to the same object,
        the resulting value is non-deterministic.
    """

    def __init__(self, value: _ValueT) -> None:
        self.value = value


class _OldImmediate(Generic[_ValueT]):
    def __init__(self, value: _ValueT) -> None:
        self.value = value


_trust_inertial = bool(int(os.environ.get("COCOTB_TRUST_INERTIAL_WRITES", "0")))

# A dictionary of pending (write_func, args), keyed by handle.
# Writes are applied oldest to newest (least recently used).
# Only the last scheduled write to a particular handle in a timestep is performed.
_write_calls: "dict[ValueObjectBase[Any, Any], Tuple[Callable[[int, Any], None], _GPISetAction, Any]]" = insertion_ordered_dict()

_write_task: Union[Task[None], None] = None

_writes_pending = Event()


async def _do_writes() -> None:
    """An internal task that schedules a ReadWrite to force writes to occur."""
    while True:
        await _writes_pending.wait()
        await ReadWrite()


def _start_write_scheduler() -> None:
    global _write_task
    if _write_task is None:
        _write_task = Task(_do_writes())
        cocotb._scheduler_inst._schedule_task(_write_task)


def _stop_write_scheduler() -> None:
    global _write_task
    if _write_task is not None:
        _write_task.cancel()
        _write_task = None
    _write_calls.clear()
    _writes_pending.clear()


def _apply_scheduled_writes() -> None:
    for func, action, value in _write_calls.values():
        func(action.value, value)
    _write_calls.clear()
    _writes_pending.clear()


if _trust_inertial:

    def _schedule_write(
        handle: "ValueObjectBase[Any, Any]",
        write_func: Callable[[int, _ValueT], None],
        action: _GPISetAction,
        value: _ValueT,
    ) -> None:
        # Trust the simulator and just write.
        write_func(action.value, value)
else:

    def _schedule_write(
        handle: "ValueObjectBase[Any, Any]",
        write_func: Callable[[int, _ValueT], None],
        action: _GPISetAction,
        value: _ValueT,
    ) -> None:
        if isinstance(current_gpi_trigger(), ReadWrite):
            # If we are already in the ReadWrite phase, apply writes immediately as an optimization.
            write_func(action.value, value)
        elif action is _GPISetAction.DEPOSIT:
            # Queue write for the beginning of the next ReadWrite phase because we can't trust the simulator. =(
            if handle in _write_calls:
                del _write_calls[handle]
            _write_calls[handle] = (write_func, action, value)
            _writes_pending.set()
        else:
            # If we are writing anything that isn't an inertial write, it must be applied immediately.
            write_func(action.value, value)


#: Type returned by the :attr:`~ValueObjectBase.value` getter and returned by the :meth:`~ValueObjectBase.get` method.
ValueGetT = TypeVar("ValueGetT")


#: Type accepted by the :attr:`~ValueObjectBase.value` setter and the :meth:`~ValueObjectBase.set` and :meth:`~ValueObjectBase.setimmediatevalue` methods.
ValueSetT = TypeVar("ValueSetT")


class ValueObjectBase(SimHandleBase, Generic[ValueGetT, ValueSetT]):
    """Abstract base class for simulation objects that have a value.

    Inherits from :class:`SimHandleBase`.
    """

    @property
    def value(self) -> ValueGetT:
        """Get or set the value of the simulation object.

        :getter: Return the current value of the simulation object.

        :setter:
            Set the value of the simulation object.

            See :class:`Deposit`, :class:`Force`, :class:`Freeze`, :class:`Release`, and :class:`Immediate`
            for additional actions that can be taken when setting a value.
            These are used like so:

            .. code-block:: python

                dut.handle.value = 1  # default Deposit action
                dut.handle.value = Deposit(2)
                dut.handle.value = Force(3)
                dut.handle.value = Freeze()
                dut.handle.value = Release()
                dut.handle.value = Immediate(4)
        """
        return self.get()

    @value.setter
    def value(self, value: ValueSetT) -> None:
        self.set(value)

    @abstractmethod
    def get(self) -> ValueGetT:
        """Return the current value of the simulation object."""

    def set(
        self,
        value: Union[
            ValueSetT,
            Deposit[ValueSetT],
            Force[ValueSetT],
            Freeze,
            Release,
            Immediate[ValueSetT],
        ],
    ) -> None:
        """Set the value of the simulation object.

        See :class:`Deposit`, :class:`Force`, :class:`Freeze`, :class:`Release`, and :class:`Immediate`
        for additional actions that can be taken when setting a value.

        Args:
            value: The value to set the simulation object to. This may include type conversion.

        Raises:
            TypeError: If the *value* is of a type that cannot be converted to a simulation value,
                or if the simulation object is immutable.
            ValueError: If the *value* is of the correct type, but the value fails to convert.
        """
        if isinstance(current_gpi_trigger(), ReadOnly):
            raise RuntimeError("Attempting settings a value during the ReadOnly phase.")
        if self.is_const:
            raise TypeError("Attempted setting an immutable object")
        if isinstance(value, Deposit):
            self._set_value(value.value, _GPISetAction.DEPOSIT)
        elif isinstance(value, Force):
            self._set_value(value.value, _GPISetAction.FORCE)
        elif isinstance(value, Freeze):
            # We assume that ValueSetT >= ValueGetT
            self._set_value(cast("ValueSetT", self.get()), _GPISetAction.FORCE)
        elif isinstance(value, Release):
            # We assume that ValueSetT >= ValueGetT
            self._set_value(cast("ValueSetT", self.get()), _GPISetAction.RELEASE)
        elif isinstance(value, Immediate):
            self._set_value(value.value, _GPISetAction.NO_DELAY)
        elif isinstance(value, _OldImmediate):
            self._set_value(value.value, _GPISetAction.OLD_IMMEDIATE)
        else:
            self._set_value(value, _GPISetAction.DEPOSIT)

    @deprecated(
        "Use `handle.set(Immediate(...))` or `handle.value = Immediate(...)` instead."
    )
    def setimmediatevalue(
        self,
        value: Union[
            ValueSetT,
            Deposit[ValueSetT],
            Force[ValueSetT],
            Freeze,
            Release,
            Immediate[ValueSetT],
        ],
    ) -> None:
        r"""Set the value of the simulation object immediately.

        See :class:`Deposit`, :class:`Force`, :class:`Freeze`, :class:`Release`, and :class:`Immediate`
        for additional actions that can be taken when setting a value.

        Passing :class:`Deposit`\ s and unwrapped values is equivalent to passing an :class:`Immediate` to :meth:`set`.

        .. deprecated:: 2.0
            Use ``handle.set(Immediate(...))`` or ``handle.value = Immediate(...)`` instead.
            This could result in a change in behavior because prior to version 2.0 this function did not set values immediately.
        """
        if isinstance(value, Deposit):
            value = _OldImmediate(value.value)  # type: ignore
        elif not isinstance(value, (Force, Freeze, Release, Immediate)):
            value = _OldImmediate(value)  # type: ignore
        self.set(value)

    @cached_property
    def is_const(self) -> bool:
        """``True`` if the simulator object is immutable, e.g. a Verilog parameter or VHDL constant or generic."""
        return self._handle.get_const()

    @abstractmethod
    def _set_value(
        self,
        value: ValueSetT,
        action: _GPISetAction,
    ) -> None:
        """Schedule a write of the given value to a simulator object.

        Conversion from multiple Python types into a type understood by the simulator is expected.
        This is used to implement the :attr:`value` property setter, :meth:`setimmediatevalue`, and :meth:`set`.
        Implementations can assume that handle isn't :meth:`const <is_const>`
        and the Scheduler is not in the :data:`ReadOnly <cocotb.SimPhase.READ_ONLY>` phase.

        Args:
            value: A value used to set the handle.
            action: Whether to deposit, force, or release the value on the handle.
        """


#: Type of value of each element in an :class:`ArrayObject`.
ElemValueT = TypeVar("ElemValueT")

#: Subtype of :class:`ValueObjectBase` returned when iterating or indexing a :class:`ArrayObject`.
ChildObjectT = TypeVar("ChildObjectT", bound=ValueObjectBase[Any, Any])


class ArrayObject(
    ValueObjectBase[Array[ElemValueT], Union[Array[ElemValueT], Sequence[ElemValueT]]],
    _RangeableObjectMixin,
    Generic[ElemValueT, ChildObjectT],
):
    """A simulation object that is an array of value-having simulation objects.

    Inherits from :class:`SimHandleBase` and :class:`ValueObjectBase`.

    With Verilog simulation objects, unpacked vectors are mapped to this type.
    Packed vectors are typically mapped to :class:`LogicArrayObject`.

    With VHDL simulation objects, all arrayed objects that aren't ``std_(u)logic``,
    ``sfixed``, ``ufixed``, ``unsigned``, ``signed``, and ``string`` are mapped to this type.

    These objects can be iterated over to yield child objects:

    .. code-block:: python

        for child in dut.array_object:
            print(child._path)

    A particular child can be retrieved using its index:

    .. code-block:: python

        child = dut.array_object[0]

        # reversed iteration over children
        for child_idx in reversed(dut.array_object.range):
            dut.array_object[child_idx]
    """

    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        super().__init__(handle, path)
        self._sub_handles: Dict[int, ChildObjectT] = {}

    def get(self) -> Array[ElemValueT]:
        """Return the current value as an :class:`~cocotb.types.Array`.

        Given the HDL array ``arr``, getting the value is equivalent to:

        +--------------+---------------------+--------------------------------------------------------------------------------------------------+
        | Verilog      | VHDL                | ``arr.get()`` is equivalent to                                                                   |
        +==============+=====================+==================================================================================================+
        | ``arr[4:7]`` | ``arr(4 to 7)``     | ``Array([arr[4].value, arr[5].value, arr[6].value, arr[7].value], range=Range(4, 'to', 7))``     |
        +--------------+---------------------+--------------------------------------------------------------------------------------------------+
        | ``arr[7:4]`` | ``arr(7 downto 4)`` | ``Array([arr[7].value, arr[6].value, arr[5].value, arr[4].value], range=Range(7, 'downto', 4))`` |
        +--------------+---------------------+--------------------------------------------------------------------------------------------------+
        """
        r = self.range
        return Array._from_handle(
            value=[self[i].value for i in r],
            range=r,
            warn_indexing=indexing_changed(r) if do_indexing_changed_warning else False,
        )

    def set(
        self,
        value: Union[
            Union[Array[ElemValueT], Sequence[ElemValueT]],
            Deposit[Union[Array[ElemValueT], Sequence[ElemValueT]]],
            Force[Union[Array[ElemValueT], Sequence[ElemValueT]]],
            Freeze,
            Release,
            Immediate[Union[Array[ElemValueT], Sequence[ElemValueT]]],
        ],
    ) -> None:
        """Set the value using an :class:`.Array`-like value.

        The simulation object is set, element-by-element, left-to-right, using the corresponding element of *value*.
        The indexes of *value* and the simulation object are not taken into account, only position.

        .. warning::
            Assigning a value to a sub-handle:

            - **Wrong**: ``dut.some_array.value[0] = 1`` (gets value as an Array, updates index 0, then throws it away)
            - **Correct**: ``dut.some_array[0].value = 1``

        Args:
            value: The value to set the signal to. This may include type conversion.

        Raises:
            TypeError: If *value* is of a type that can't be assigned to the simulation object.

        .. warning::
            Exceptions from array element :meth:`.ValueObjectBase.set` calls will be propagated up,
            so the actual set of exceptions possible is greater than this list.
        """
        super().set(value)

    def _set_value(
        self,
        value: Union[Array[ElemValueT], Sequence[ElemValueT]],
        action: _GPISetAction,
    ) -> None:
        if len(value) != len(self):
            raise ValueError(
                f"Assigning list of length {len(value)} to object {self._name} of length {len(self)}"
            )
        for elem, self_idx in zip(value, self.range):
            self[self_idx]._set_value(elem, action)

    def __getitem__(self, index: int) -> ChildObjectT:
        if isinstance(index, slice):
            raise TypeError("Slicing is not supported")
        if index in self._sub_handles:
            return self._sub_handles[index]
        new_handle = self._handle.get_handle_by_index(index)
        if not new_handle:
            raise IndexError(f"{self._path} contains no object at index {index}")
        path = self._path + "[" + str(index) + "]"
        self._sub_handles[index] = cast(
            "ChildObjectT", _make_sim_object(new_handle, path)
        )
        return self._sub_handles[index]

    def __iter__(self) -> Iterable[ChildObjectT]:
        for i in self.range:
            yield self[i]


class _NonIndexableValueObjectBase(ValueObjectBase[ValueGetT, ValueSetT]):
    """ValueObject that is treated as a single object in the GPI.

    NonArrayValueObjects support :meth:`value_change` triggers.
    """

    @cached_property
    def value_change(self) -> ValueChange:
        """A trigger which fires whenever the value changes."""
        if self.is_const:
            raise TypeError("Can't get ValueChange on immutable signal.")
        return ValueChange._make(self)

    @cached_property
    def _edge(self) -> Edge:
        if self.is_const:
            raise TypeError("Can't get Edge on immutable signal.")
        return Edge._make(self)


class LogicObject(_NonIndexableValueObjectBase[Logic, Union[Logic, int, str]]):
    """A scalar logic simulation object.

    Inherits from :class:`SimHandleBase` and :class:`ValueObjectBase`.

    Verilog data types that map to this object:

        * ``logic``
        * ``bit``

    VHDL types that map to this object:

        * ``std_logic``
        * ``std_ulogic``
        * ``bit``
    """

    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        super().__init__(handle, path)

    def _set_value(
        self,
        value: Union[Logic, int, str],
        action: _GPISetAction,
    ) -> None:
        value_: str
        if isinstance(value, (int, str)):
            value_ = str(Logic(value))

        elif isinstance(value, LogicArray):
            if len(value) != 1:
                raise ValueError(
                    f"Cannot assign value of length {len(value)} to handle of length 1"
                )
            value_ = str(value)

        elif isinstance(value, Logic):
            value_ = str(value)

        else:
            raise TypeError(
                f"Unsupported type for value assignment: {type(value)} ({value!r})"
            )

        _schedule_write(self, self._handle.set_signal_val_binstr, action, value_)

    def get(self) -> Logic:
        """Return the current value of the simulation object as a :class:`.Logic`."""
        binstr = self._handle.get_signal_val_binstr()
        return Logic(binstr)

    def set(
        self,
        value: Union[
            Union[Logic, int, str],
            Deposit[Union[Logic, int, str]],
            Force[Union[Logic, int, str]],
            Freeze,
            Release,
            Immediate[Union[Logic, int, str]],
        ],
    ) -> None:
        """Set the value of the simulation object using a :class:`.Logic`-like value.

        Args:
            value: The value to set the simulation object to.

        Raises:
            TypeError: If *value* is of a type that can't be assigned to the simulation object, or readily converted into a type that can.
            ValueError: If *value* would not fit in the bounds of the simulation object.
        """
        super().set(value)

    @cached_property
    def rising_edge(self) -> RisingEdge:
        """A trigger which fires whenever the value changes to a ``1``."""
        if self.is_const:
            raise TypeError("Can't get RisingEdge on immutable signal")
        return RisingEdge._make(self)

    @cached_property
    def falling_edge(self) -> FallingEdge:
        """A trigger which fires whenever the value changes to a ``0``."""
        if self.is_const:
            raise TypeError("Can't get FallingEdge on immutable signal")
        return FallingEdge._make(self)

    def __len__(self) -> int:
        return 1

    @deprecated(
        "`int(handle)` casts have been deprecated. Use `int(handle.value)` instead."
    )
    def __int__(self) -> int:
        return int(self.value)

    @deprecated(
        "`str(handle)` casts have been deprecated. Use `str(handle.value)` instead."
    )
    def __str__(self) -> str:
        return str(self.value)


class LogicArrayObject(
    _NonIndexableValueObjectBase[LogicArray, Union[LogicArray, Logic, int, str]],
    _RangeableObjectMixin,
):
    """A logic array simulation object.

    Inherits from :class:`SimHandleBase` and :class:`ValueObjectBase`.

    Verilog types that map to this object:

        * packed any-dimensional vectors of ``logic`` or ``bit``
        * packed any-dimensional vectors of packed structures

    VHDL types that map to this object:

        * ``std_logic_vector`` and ``std_ulogic_vector``
        * ``unsigned``
        * ``signed``
        * ``ufixed``
        * ``sfixed``
        * ``float``
    """

    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        super().__init__(handle, path)

    def _set_value(
        self,
        value: Union[LogicArray, Logic, int, str],
        action: _GPISetAction,
    ) -> None:
        value_: str
        if isinstance(value, int):
            min_val, max_val = _value_limits(len(self), _Limits.VECTOR_NBIT)
            if min_val <= value <= max_val:
                if len(self) <= 32:
                    _schedule_write(
                        self, self._handle.set_signal_val_int, action, value
                    )
                    return

                # LogicArray used for checking
                if value < 0:
                    value_ = str(
                        LogicArray.from_signed(
                            value,
                            Range(len(self) - 1, "downto", 0),
                        )
                    )
                else:
                    value_ = str(
                        LogicArray.from_unsigned(
                            value,
                            Range(len(self) - 1, "downto", 0),
                        )
                    )
            else:
                raise ValueError(
                    f"Int value ({value!r}) out of range for assignment of {len(self)!r}-bit signal ({self._name!r})"
                )

        elif isinstance(value, str):
            # LogicArray used for checking
            value_ = str(LogicArray(value, self.range))

        elif isinstance(value, LogicArray):
            if len(self) != len(value):
                raise ValueError(
                    f"cannot assign value of length {len(value)} to handle of length {len(self)}"
                )
            value_ = str(value)

        elif isinstance(value, Logic):
            if len(self) != 1:
                raise ValueError(
                    f"cannot assign value of length 1 to handle of length {len(self)}"
                )
            value_ = str(value)

        else:
            raise TypeError(
                f"Unsupported type for value assignment: {type(value)} ({value!r})"
            )

        _schedule_write(self, self._handle.set_signal_val_binstr, action, value_)

    def get(self) -> LogicArray:
        """Return the current value of the simulation object as a :class:`.LogicArray`."""
        binstr = self._handle.get_signal_val_binstr()
        return LogicArray._from_handle(
            value=binstr,
            warn_indexing=indexing_changed(self.range)
            if do_indexing_changed_warning
            else False,
        )

    def set(
        self,
        value: Union[
            Union[LogicArray, Logic, int, str],
            Deposit[Union[LogicArray, Logic, int, str]],
            Force[Union[LogicArray, Logic, int, str]],
            Freeze,
            Release,
            Immediate[Union[LogicArray, Logic, int, str]],
        ],
    ) -> None:
        """Set the value of the simulation object using a :class:`.LogicArray`-like value.

        Args:
            value: The value to set the simulation object to.

        Raises:
            TypeError: If *value* is of a type that can't be assigned to the simulation object, or readily converted into a type that can.
            ValueError: If *value* would not fit in the bounds of the simulation object.

        .. versionchanged:: 2.0
            Using :class:`ctypes.Structure` objects to set values was removed.
            Convert the struct object to a :class:`~cocotb.types.LogicArray` before assignment using
            ``LogicArray("".join(format(int(byte), "08b") for byte in bytes(struct_obj)))`` instead.

        .. versionchanged:: 2.0
            Using :class:`dict` objects to set values was removed.
            Convert the dictionary to an integer before assignment using
            ``sum(v << (d['bits'] * i) for i, v in enumerate(d['values']))`` instead.

        .. versionchanged:: 2.0
            Supplying too large of an :class:`int` value results in raising a :exc:`ValueError` instead of an :exc:`OverflowError`.
        """
        super().set(value)

    @deprecated(
        "`int(handle)` casts have been deprecated. Use `int(handle.value)` instead."
    )
    def __int__(self) -> int:
        return int(self.value)

    @deprecated(
        "`str(handle)` casts have been deprecated. Use `str(handle.value)` instead."
    )
    def __str__(self) -> str:
        return str(self.value)

    def __len__(self) -> int:
        # can't use `range` to get length because `range` is for outer-most dimension only
        # and this object needs to support multi-dimensional packed arrays.
        return self._len

    @cached_property
    def _len(self) -> int:
        return self._handle.get_num_elems()

    def __getitem__(self, _: object) -> NoReturn:
        raise TypeError(
            "Packed objects, either arrays or structs, cannot be indexed.\n"
            "Try instead reading the whole value and slicing: `t = handle.value; t[0:3]`.\n"
            "If you need to use an element in an Edge Trigger, consider making the array or struct unpacked.\n"
            "Alternatively, use `ValueChange` on the whole object and check the bit(s) you care about for changes afterwards."
        )


class RealObject(_NonIndexableValueObjectBase[float, float]):
    """A floating point simulation object.

    Inherits from :class:`SimHandleBase` and :class:`ValueObjectBase`.

    This type is used when a ``real`` object in VHDL or ``float`` object in Verilog is seen.
    """

    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        super().__init__(handle, path)

    def _set_value(
        self,
        value: float,
        action: _GPISetAction,
    ) -> None:
        if not isinstance(value, (float, int)):
            raise TypeError(
                f"Unsupported type for real value assignment: {type(value)} ({value!r})"
            )

        _schedule_write(self, self._handle.set_signal_val_real, action, value)

    def get(self) -> float:
        """Return the current value of the simulation object as a :class:`float`."""
        return self._handle.get_signal_val_real()

    def set(
        self,
        value: Union[
            float,
            Deposit[float],
            Force[float],
            Freeze,
            Release,
            Immediate[float],
        ],
    ) -> None:
        """Set the value of the simulation object using a :class:`float` value.

        Args:
            value: The value to set the simulation object to.

        Raises:
            TypeError: If *value* is any type other than :class:`float`.
        """
        super().set(value)

    @deprecated(
        "`float(handle)` casts have been deprecated. Use `float(handle.value)` instead."
    )
    def __float__(self) -> float:
        return self.value


class EnumObject(_NonIndexableValueObjectBase[int, int]):
    """An enumeration simulation object.

    Inherits from :class:`SimHandleBase` and :class:`ValueObjectBase`.

    This type is used when an enumerated-type simulation object is seen that aren't a "logic" or similar type.
    The value of this object is represented with an :class:`int`.

    For VHDL objects, the value being represented is the enumeration value at the integer index into the original ``type`` declaration,
    as if it were a 1-based array.

    For Verilog objects, enumerations are little more than named integer values.
    There may be many enumeration values that a given :class:`int` value represents.
    """

    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        super().__init__(handle, path)

    def _set_value(
        self,
        value: int,
        action: _GPISetAction,
    ) -> None:
        if not isinstance(value, int):
            raise TypeError(
                f"Unsupported type for enum value assignment: {type(value)} ({value!r})"
            )

        min_val, max_val = _value_limits(32, _Limits.UNSIGNED_NBIT)
        if min_val <= value <= max_val:
            _schedule_write(self, self._handle.set_signal_val_int, action, value)
        else:
            raise ValueError(
                f"Int value ({value!r}) out of range for assignment of enum signal ({self._name!r})"
            )

    def get(self) -> int:
        """Return the current value of the simulation object as an :class:`int`.

        See :class:`EnumObject` for details on what :class:`int` values correspond to which enumeration values.
        """
        return self._handle.get_signal_val_long()

    def set(
        self,
        value: Union[
            int,
            Deposit[int],
            Force[int],
            Freeze,
            Release,
            Immediate[int],
        ],
    ) -> None:
        """Set the value of the simulation object using an :class:`int`.

        See :class:`EnumObject` for details on what :class:`int` values correspond to which enumeration values.

        Args:
            value: The value to set the simulation object to.

        Raises:
            TypeError: If *value* is any type other than :class:`int`.
            ValueError: If *value* would not fit in a 32-bit signed integer.

        .. versionchanged:: 2.0
            Supplying too large of a value results in raising a :exc:`ValueError` instead of an :exc:`OverflowError`.
        """
        super().set(value)

    @deprecated(
        "`int(handle)` casts have been deprecated. Use `int(handle.value)` instead."
    )
    def __int__(self) -> int:
        return int(self.value)


class IntegerObject(_NonIndexableValueObjectBase[int, int]):
    """An integer simulation object.

    Inherits from :class:`SimHandleBase` and :class:`ValueObjectBase`.

    Verilog types that map to this object:

        * ``byte``
        * ``shortint``
        * ``int``
        * ``longint``

    This type should not be used for the 4-state integer types ``integer`` and ``time``.

    VHDL types that map to this object:

        * ``integer``
        * ``natural``
        * ``positive``

    Objects that use this type are assumed to be two's complement 32-bit integers with 2-state (``0`` and ``1``) bits.
    """

    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        super().__init__(handle, path)

    def _set_value(
        self,
        value: int,
        action: _GPISetAction,
    ) -> None:
        if not isinstance(value, int):
            raise TypeError(
                f"Unsupported type for integer value assignment: {type(value)} ({value!r})"
            )

        min_val, max_val = _value_limits(32, _Limits.SIGNED_NBIT)
        if min_val <= value <= max_val:
            _schedule_write(self, self._handle.set_signal_val_int, action, value)
        else:
            raise ValueError(
                f"Int value ({value!r}) out of range for assignment of integer signal ({self._name!r})"
            )

    def get(self) -> int:
        """Return the current value of the simulation object as an :class:`int`."""
        return self._handle.get_signal_val_long()

    def set(
        self,
        value: Union[
            int,
            Deposit[int],
            Force[int],
            Freeze,
            Release,
            Immediate[int],
        ],
    ) -> None:
        """Set the the value of the simulation object using an :class:`int` value.

        Args:
            value: The value to set the simulation object to.

        Raises:
            TypeError: If *value* is any type other than :class:`int`.
            ValueError: If *value* would not fit in a 32-bit signed integer.

        .. versionchanged:: 2.0
            Supplying too large of a value results in raising a :exc:`ValueError` instead of an :exc:`OverflowError`.
        """
        super().set(value)

    @deprecated(
        "`int(handle)` casts have been deprecated. Use `int(handle.value)` instead."
    )
    def __int__(self) -> int:
        return self.value


class StringObject(
    _NonIndexableValueObjectBase[bytes, bytes],
    _RangeableObjectMixin,
):
    """A string simulation object.

    Inherits from :class:`SimHandleBase` and :class:`ValueObjectBase`.

    This type is used when a ``string`` (VHDL or Verilog) simulation object is seen.
    """

    def __init__(self, handle: simulator.gpi_sim_hdl, path: Optional[str]) -> None:
        super().__init__(handle, path)

    def _set_value(
        self,
        value: bytes,
        action: _GPISetAction,
    ) -> None:
        if not isinstance(value, (bytes, bytearray)):
            raise TypeError(
                f"Unsupported type for string value assignment: {type(value)} ({value!r})"
            )
        _schedule_write(self, self._handle.set_signal_val_str, action, value)

    def get(self) -> bytes:
        """Return the current value of the simulation object as a :class:`bytes`."""
        return self._handle.get_signal_val_str()

    def set(
        self,
        value: Union[
            bytes,
            Deposit[bytes],
            Force[bytes],
            Freeze,
            Release,
            Immediate[bytes],
        ],
    ) -> None:
        """Set the value of the simulation object with a :class:`bytes` or :class:`bytearray` value.

        When *value*'s length is less than the simulation object's,
        the value is padded with NUL (``'\0'``) characters up to the appropriate length.
        When the value's length is greater than the simulation object's,
        the value is truncated without a NUL terminator to the appropriate length,
        without warning.

        Strings in both Verilog and VHDL are byte arrays without any particular encoding.
        Encoding must be done to turn Python strings into byte arrays.
        Because `there are many encodings <https://docs.python.org/3/library/codecs.html#standard-encodings>`_,
        this step is left up to the user.

        An example of how encoding and decoding could be accomplished using an ASCII string.

        .. code-block:: python

            # lowercase a string
            value = dut.string_handle.value.decode("ascii")
            value = value.lower()
            dut.string_handle.value = value.encode("ascii")

        Args:
            value: The value to set the simulation object to.

        Raises:
            TypeError: If *value* is any type other than :class:`bytes`.

        .. versionchanged:: 1.4
            Takes :class:`bytes` instead of :class:`str`.
            Users are now expected to choose an encoding when using these objects.
        """
        super().set(value)

    @deprecated(
        '`str(handle)` casts have been deprecated. Use `handle.value.decode("ascii")` instead.'
    )
    def __str__(self) -> str:
        return self.value.decode("ascii")


_ConcreteHandleTypes = Union[
    HierarchyObject,
    HierarchyArrayObject[SimHandleBase],
    LogicObject,
    LogicArrayObject,
    ArrayObject[Any, ValueObjectBase[Any, Any]],
    RealObject,
    IntegerObject,
    EnumObject,
    StringObject,
]


_handle2obj: Dict[
    simulator.gpi_sim_hdl,
    _ConcreteHandleTypes,
] = {}

_type2cls: Dict[int, Type[_ConcreteHandleTypes]] = {
    simulator.MODULE: HierarchyObject,
    simulator.STRUCTURE: HierarchyObject,
    simulator.PACKED_STRUCTURE: LogicArrayObject,
    simulator.LOGIC: LogicObject,
    simulator.LOGIC_ARRAY: LogicArrayObject,
    simulator.NETARRAY: ArrayObject[Any, ValueObjectBase[Any, Any]],
    simulator.REAL: RealObject,
    simulator.INTEGER: IntegerObject,
    simulator.ENUM: EnumObject,
    simulator.STRING: StringObject,
    simulator.GENARRAY: HierarchyArrayObject[SimHandleBase],
    simulator.PACKAGE: HierarchyObject,
}


def _make_sim_object(
    handle: simulator.gpi_sim_hdl, path: Optional[str] = None
) -> SimHandleBase:
    """Factory function to create the correct type of `SimHandle` object.

    Args:
        handle: The GPI handle to the simulator object.
        path: Path to this handle.

    Returns:
        An appropriate :class:`SimHandleBase` object.

    Raises:
        NotImplementedError: If no matching object for GPI type could be found.
    """

    # Enforce singletons since it's possible to retrieve handles avoiding
    # the hierarchy by getting driver/load information
    try:
        return _handle2obj[handle]
    except KeyError:
        pass

    t = handle.get_type()
    if t not in _type2cls:
        raise NotImplementedError(
            f"Couldn't find a matching object for GPI type {handle.get_type_string()}({t}) (path={path})"
        )
    obj = _type2cls[t](handle, path)
    _handle2obj[handle] = obj
    return obj



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import os


def start_cocotb_library_coverage(_: object) -> None:  # pragma: no cover
    if "COCOTB_LIBRARY_COVERAGE" not in os.environ:
        return
    try:
        import coverage  # noqa: PLC0415
    except (ImportError, ModuleNotFoundError):
        raise RuntimeError(
            "cocotb library coverage collection requested but coverage package not available. Install it using `pip install coverage`."
        ) from None
    else:
        library_coverage = coverage.coverage(
            data_file=".coverage.cocotb",
            config_file=False,
            branch=True,
            source=["cocotb"],
        )
        library_coverage.start()

        def stop_library_coverage() -> None:
            library_coverage.stop()
            library_coverage.save()  # pragma: no cover

        # This must come after `library_coverage.start()` to ensure coverage is being
        # collected on the cocotb library before importing from it.
        from cocotb._init import _register_shutdown_callback  # noqa: PLC0415

        _register_shutdown_callback(stop_library_coverage)



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

import functools
import inspect
from enum import Enum
from itertools import product
from typing import (
    Callable,
    Coroutine,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from cocotb._base_triggers import Trigger
from cocotb._py_compat import Protocol, TypeAlias
from cocotb._typing import TimeUnit


class Test:
    """A cocotb test in a regression.

    Args:
        func:
            The test function object.

        name:
            The name of the test function.
            Defaults to ``func.__qualname__`` (the dotted path to the test function in the module).

        module:
            The name of the module containing the test function.
            Defaults to ``func.__module__`` (the name of the module containing the test function).

        doc:
            The docstring for the test.
            Defaults to ``func.__doc__`` (the docstring of the test function).

        timeout_time:
            Simulation time duration before the test is forced to fail with a :exc:`~cocotb.triggers.SimTimeoutError`.

        timeout_unit:
            Unit of ``timeout_time``, accepts any unit that :class:`~cocotb.triggers.Timer` does.

        expect_fail:
            If ``True`` and the test fails a functional check via an :keyword:`assert` statement, :func:`pytest.raises`,
            :func:`pytest.warns`, or :func:`pytest.deprecated_call`, the test is considered to have passed.
            If ``True`` and the test passes successfully, the test is considered to have failed.

        expect_error:
            Mark the result as a pass only if one of the given exception types is raised in the test.

        skip:
            Don't execute this test as part of the regression.
            The test can still be run manually by setting :envvar:`COCOTB_TESTCASE`.

        stage:
            Order tests logically into stages.
            Tests from earlier stages are run before tests from later stages.
    """

    def __init__(
        self,
        *,
        func: Callable[..., Coroutine[Trigger, None, None]],
        name: Optional[str] = None,
        module: Optional[str] = None,
        doc: Optional[str] = None,
        timeout_time: Optional[float] = None,
        timeout_unit: TimeUnit = "step",
        expect_fail: bool = False,
        expect_error: Union[Type[BaseException], Tuple[Type[BaseException], ...]] = (),
        skip: bool = False,
        stage: int = 0,
    ) -> None:
        self.func: Callable[..., Coroutine[Trigger, None, None]] = func
        self.timeout_time = timeout_time
        self.timeout_unit = timeout_unit
        self.expect_fail = expect_fail
        if isinstance(expect_error, type):
            expect_error = (expect_error,)
        self.expect_error = expect_error
        self.skip = skip
        self.stage = stage
        self.name = self.func.__qualname__ if name is None else name
        self.module = self.func.__module__ if module is None else module
        self.doc = self.func.__doc__ if doc is None else doc
        if self.doc is not None:
            # cleanup docstring using `trim` function from PEP257
            self.doc = inspect.cleandoc(self.doc)
        self.fullname = f"{self.module}.{self.name}"


TestFuncType: TypeAlias = Callable[..., Coroutine[Trigger, None, None]]


class Parameterized:
    def __init__(
        self,
        test_function: TestFuncType,
        options: List[
            Union[
                Tuple[str, Sequence[object]],
                Tuple[Sequence[str], Sequence[Sequence[object]]],
            ]
        ],
    ) -> None:
        self.test_template = Test(func=test_function)
        self.options = options
        # we are assuming the input checking is done in parametrize()

        self._option_reprs: Dict[str, List[str]] = {}

        for name, values in options:
            if isinstance(name, str):
                self._option_reprs[name] = _reprs(values)
            else:
                # transform to Dict[name, values]
                transformed: Dict[str, List[object]] = {}
                for nam_idx, nam in enumerate(name):
                    transformed[nam] = []
                    for value_array in cast("Sequence[Sequence[object]]", values):
                        value = value_array[nam_idx]
                        transformed[nam].append(value)
                for n, vs in transformed.items():
                    self._option_reprs[n] = _reprs(vs)

    def generate_tests(self) -> Iterable[Test]:
        test_func = self.test_template.func
        test_func_name = self.test_template.name

        # this value is a list of ranges of the same length as each set of values in self.options for passing to itertools.product
        option_indexes = [range(len(option[1])) for option in self.options]

        # go through the cartesian product of all values of all options
        for selected_options in product(*option_indexes):
            test_kwargs: Dict[str, object] = {}
            test_name_pieces: List[str] = [test_func_name]
            for option_idx, select_idx in enumerate(selected_options):
                option_name, option_values = self.options[option_idx]
                selected_value = option_values[select_idx]

                if isinstance(option_name, str):
                    # single params per option
                    selected_value = cast("Sequence[object]", selected_value)
                    test_kwargs[option_name] = selected_value
                    test_name_pieces.append(
                        f"/{option_name}={self._option_reprs[option_name][select_idx]}"
                    )
                else:
                    # multiple params per option
                    selected_value = cast("Sequence[object]", selected_value)
                    for n, v in zip(option_name, selected_value):
                        test_kwargs[n] = v
                        test_name_pieces.append(
                            f"/{n}={self._option_reprs[n][select_idx]}"
                        )

            parametrized_test_name = "".join(test_name_pieces)

            # create wrapper function to bind kwargs
            @functools.wraps(test_func)
            async def _my_test(
                dut: object, kwargs: Dict[str, object] = test_kwargs
            ) -> None:
                await test_func(dut, **kwargs)

            yield Test(
                func=_my_test,
                name=parametrized_test_name,
                timeout_time=self.test_template.timeout_time,
                timeout_unit=self.test_template.timeout_unit,
                expect_fail=self.test_template.expect_fail,
                expect_error=self.test_template.expect_error,
                skip=self.test_template.skip,
                stage=self.test_template.stage,
            )


def _reprs(values: Sequence[object]) -> List[str]:
    result: List[str] = []
    for value in values:
        value_repr = _repr(value)
        if value_repr is None:
            # non-representable value in option, so default to index strings and give up
            return [str(i) for i in range(len(values))]
        else:
            result.append(value_repr)
    return result


def _repr(v: object) -> Optional[str]:
    if isinstance(v, Enum):
        return v.name
    elif isinstance(v, str):
        if len(v) <= 10 and v.isidentifier():
            return v
        else:
            return None
    elif isinstance(v, (int, float, bool, type(None))):
        return repr(v)
    elif isinstance(v, type):
        return v.__qualname__
    elif hasattr(v, "__qualname__"):
        return v.__qualname__
    else:
        return None


class TestDecoratorType(Protocol):
    # TODO use position only argument for *obj* so we aren't tied to that name.
    @overload
    def __call__(self, obj: TestFuncType) -> Test: ...
    @overload
    def __call__(self, obj: Parameterized) -> Parameterized: ...


@overload
def test(obj: TestFuncType) -> Test: ...


@overload
def test(obj: Parameterized) -> Parameterized: ...


@overload
def test(
    *,
    timeout_time: Optional[float] = None,
    timeout_unit: TimeUnit = "step",
    expect_fail: bool = False,
    expect_error: Union[Type[BaseException], Tuple[Type[BaseException], ...]] = (),
    skip: bool = False,
    stage: int = 0,
    name: Optional[str] = None,
) -> TestDecoratorType: ...


def test(
    obj: Optional[Union[TestFuncType, Parameterized]] = None,
    *,
    timeout_time: Optional[float] = None,
    timeout_unit: TimeUnit = "step",
    expect_fail: bool = False,
    expect_error: Union[Type[BaseException], Tuple[Type[BaseException], ...]] = (),
    skip: bool = False,
    stage: int = 0,
    name: Optional[str] = None,
) -> Union[
    Test,
    Parameterized,
    TestDecoratorType,
]:
    r"""
    Decorator to register a Callable which returns a Coroutine as a test.

    The test decorator provides a test timeout, and allows us to mark tests as skipped or expecting errors or failures.
    Tests are evaluated in the order they are defined in a test module.

    Usage:
        .. code-block:: python

            @cocotb.test(timeout_time=10, timeout_unit="ms")
            async def test_thing(dut): ...

    Args:
        timeout_time:
            Simulation time duration before timeout occurs.

            .. versionadded:: 1.3

            .. note::
                Test timeout is intended for protection against deadlock.
                Users should use :class:`~cocotb.triggers.with_timeout` if they require a
                more general-purpose timeout mechanism.

        timeout_unit:
            Units of timeout_time, accepts any units that :class:`~cocotb.triggers.Timer` does.

            .. versionadded:: 1.3

            .. versionchanged:: 2.0
                Passing ``None`` as the *timeout_unit* argument was removed, use ``'step'`` instead.

        expect_fail:
            If ``True`` and the test fails a functional check via an :keyword:`assert` statement, :class:`pytest.raises`,
            :class:`pytest.warns`, or :class:`pytest.deprecated_call` the test is considered to have passed.
            If ``True`` and the test passes successfully, the test is considered to have failed.

        expect_error:
            Mark the result as a pass only if one of the exception types is raised in the test.
            This is primarily for cocotb internal regression use for when a simulator error is expected.

            Users are encouraged to use the following idiom instead::

                @cocotb.test()
                async def my_test(dut):
                    try:
                        await thing_that_should_fail()
                    except ExceptionIExpect:
                        pass
                    else:
                        assert False, "Exception did not occur"

            .. versionchanged:: 1.3
                Specific exception types can be expected

            .. versionchanged:: 2.0
                Passing a :class:`bool` value was removed.
                Pass a specific :class:`Exception` or a tuple of Exceptions instead.

        skip:
            Don't execute this test as part of the regression. Test can still be run
            manually by setting :make:var:`COCOTB_TESTCASE`.

        stage:
            Order tests logically into stages, where multiple tests can share a stage.
            Defaults to 0.

        name:
            Override the default name of the test.
            The default test name is the :meth:`__qualname__` of the decorated test function.

            .. versionadded:: 2.0

    Returns:
        The test function to which the decorator is applied.

    .. note::

        To extend the test decorator, use the following template to create a new
        ``cocotb.test``\-like wrapper.

        .. code-block:: python

            import functools


            def test_extender(**decorator_kwargs):
                def decorator(obj):
                    @cocotb.test(**decorator_kwargs)
                    @functools.wraps(obj)
                    async def test(dut, **test_kwargs):
                        # your code here
                        ...

                    return obj

                return decorator

    .. versionchanged:: 2.0
        Support using decorator on test function without supplying parameters first.

        Assumes all default values for the test parameters.

        .. code-block:: python

            @cocotb.test
            async def test_thing(dut): ...


    .. versionchanged:: 2.0
        Decorated tests now return the decorated object.

    """
    if obj is not None:
        if isinstance(obj, Parameterized):
            return obj
        else:
            return Test(func=obj)

    @overload
    def wrapper(obj: TestFuncType) -> Test: ...

    @overload
    def wrapper(obj: Parameterized) -> Parameterized: ...

    def wrapper(obj: Union[TestFuncType, Parameterized]) -> Union[Test, Parameterized]:
        if isinstance(obj, Parameterized):
            obj.test_template = Test(
                func=obj.test_template.func,
                name=name,
                timeout_time=timeout_time,
                timeout_unit=timeout_unit,
                expect_fail=expect_fail,
                expect_error=expect_error,
                skip=skip,
                stage=stage,
            )
            return obj
        else:
            return Test(
                func=obj,
                name=name,
                timeout_time=timeout_time,
                timeout_unit=timeout_unit,
                expect_fail=expect_fail,
                expect_error=expect_error,
                skip=skip,
                stage=stage,
            )

    return wrapper


def parametrize(
    *options_by_tuple: Union[
        Tuple[str, Sequence[object]], Tuple[Sequence[str], Sequence[Sequence[object]]]
    ],
    **options_by_name: Sequence[object],
) -> Callable[[TestFuncType], Parameterized]:
    """Decorator to generate parametrized tests from a single test function.

    Decorates a test function with named test parameters.
    The call to ``parametrize`` should include the name of each test parameter and the possible values each parameter can hold.
    This will generate a test for each of the Cartesian products of the parameters and their values.

    .. code-block:: python

        @cocotb.test(
            skip=False,
        )
        @cocotb.parametrize(
            arg1=[0, 1],
            arg2=["a", "b"],
        )
        async def my_test(arg1: int, arg2: str) -> None: ...

    The above is equivalent to the following.

    .. code-block:: python

        @cocotb.test(skip=False)
        async def my_test_0_a() -> None:
            arg1, arg2 = 0, "a"
            ...


        @cocotb.test(skip=False)
        async def my_test_0_b() -> None:
            arg1, arg2 = 0, "b"
            ...


        @cocotb.test(skip=False)
        async def my_test_1_a() -> None:
            arg1, arg2 = 1, "a"
            ...


        @cocotb.test(skip=False)
        async def my_test_1_b() -> None:
            arg1, arg2 = 1, "b"
            ...

    Options can also be specified in much the same way that :meth:`TestFactory.add_option <cocotb.regression.TestFactory.add_option>` can,
    either by supplying tuples of the parameter name to values,
    or a sequence of variable names and a sequence of values.

    .. code-block:: python

        @cocotb.parametrize(
            ("arg1", [0, 1]),
            (("arg2", "arg3"), [(1, 2), (3, 4)]),
        )
        async def my_test_2(arg1: int, arg2: int, arg3: int) -> None: ...

    Args:
        options_by_tuple:
            Tuple of parameter name to sequence of values for that parameter,
            or tuple of sequence of parameter names to sequence of sequences of values for that pack of parameters.

        options_by_name:
            Mapping of parameter name to sequence of values for that parameter.

    .. versionadded:: 2.0
    """

    # check good inputs
    for i, option_by_tuple in enumerate(options_by_tuple):
        if len(option_by_tuple) != 2:
            raise ValueError(
                f"Invalid option tuple {i}, expected exactly two fields `(name, values)`"
            )
        name, values = option_by_tuple
        if not isinstance(name, str):
            for n in name:
                if not n.isidentifier():
                    raise ValueError("Option names must be valid Python identifiers")
            values = cast("Sequence[Sequence[object]]", values)
            for value in values:
                if len(name) != len(value):
                    raise ValueError(
                        f"Invalid option tuple {i}, mismatching number of parameters ({name}) and values ({value})"
                    )
        elif not name.isidentifier():
            raise ValueError("Option names must be valid Python identifiers")

    options = [*options_by_tuple, *options_by_name.items()]

    def wrapper(f: TestFuncType) -> Parameterized:
        return Parameterized(f, options)

    return wrapper



#!/usr/bin/env python
# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""
Module for querying the cocotb configuration

This module provides information in module global variables and through a
``main()`` function that is used in the cocotb-config script.

Global variables:
    share_dir: str, path where the cocotb data is stored
    makefiles_dir: str, path where the cocotb makefiles are installed
    libs_dir: str, path where the cocotb interface libraries are located
"""

import argparse
import os
import sys
import textwrap
from pathlib import Path

import find_libpython

import cocotb_tools

base_tools_dir = Path(cocotb_tools.__file__).parent.resolve()
base_cocotb_dir = (base_tools_dir.parent / "cocotb").resolve()
if not (base_cocotb_dir.exists() and (base_cocotb_dir / "libs").exists()):
    import cocotb

    base_cocotb_dir = Path(cocotb.__file__).parent.resolve()

share_dir = base_cocotb_dir.joinpath("share")
libs_dir = base_cocotb_dir.joinpath("libs")
makefiles_dir = base_tools_dir.joinpath("makefiles")


def _get_version() -> str:
    import cocotb  # noqa: PLC0415

    return cocotb.__version__


def _help_vars_text() -> str:
    if "dev" in _get_version():
        doclink = "https://docs.cocotb.org/en/development/library_reference.html"
    else:
        doclink = f"https://docs.cocotb.org/en/v{_get_version()}/library_reference.html"

    # NOTE: make sure to keep "helpmsg" aligned with docs/source/library_reference.rst
    helpmsg = textwrap.dedent(
        """\
        The following variables are environment variables:

        cocotb
        ------
        COCOTB_TOPLEVEL           Instance in the hierarchy to use as the DUT
        COCOTB_RANDOM_SEED        Random seed, to recreate a previous test stimulus
        COCOTB_ANSI_OUTPUT        Force cocotb to print or not print in color
        COCOTB_REDUCED_LOG_FMT    Display log lines shorter
        COCOTB_ATTACH             Pause time value in seconds before the simulator start
        COCOTB_ENABLE_PROFILING   Performance analysis of the Python portion of cocotb
        COCOTB_LOG_LEVEL          Default logging level (default INFO)
        COCOTB_RESOLVE_X          How to resolve X, Z, U, W, - on integer conversion
        LIBPYTHON_LOC             Absolute path to libpython

        Regression Manager
        ------------------
        COCOTB_PDB_ON_EXCEPTION   Drop into the Python debugger (pdb) on exception
        COCOTB_TEST_MODULES       Module(s) to search for test functions (comma-separated)
        COCOTB_TESTCASE           Test function(s) to run (comma-separated list)
        COCOTB_RESULTS_FILE       File name for xUnit XML tests results
        COCOTB_USER_COVERAGE      Collect Python user coverage (HDL for some simulators)
        COVERAGE_RCFILE           Configuration for user code coverage

        GPI
        ---
        GPI_EXTRA                       Extra libraries to load at runtime (comma-separated)

        Scheduler
        ---------
        COCOTB_SCHEDULER_DEBUG         Enable additional output of coroutine scheduler
        COCOTB_TRUST_INERTIAL_WRITES   Trust inertial writes rather than mock them using scheduler

        For details, see {}"""
    ).format(doclink)
    return helpmsg


def lib_name(interface: str, simulator: str) -> str:
    """
    Return the name of interface library for given interface (VPI/VHPI/FLI) and simulator.
    """

    interface_name = interface.lower()
    supported_interfaces = ["vpi", "vhpi", "fli"]
    if interface_name not in supported_interfaces:
        raise ValueError(
            "Wrong interface used. Supported: " + ", ".join(supported_interfaces)
        )

    simulator_name = simulator.lower()
    supported_sims = [
        "icarus",
        "verilator",
        "questa",
        "modelsim",
        "ius",
        "xcelium",
        "vcs",
        "ghdl",
        "riviera",
        "activehdl",
        "cvc",
        "nvc",
        "dsim",
    ]
    if simulator not in supported_sims:
        raise ValueError(
            "Wrong simulator name. Supported: " + ", ".join(supported_sims)
        )

    if simulator_name in ["questa", "cvc"]:
        library_name = "modelsim"
    elif simulator_name == "xcelium":
        library_name = "ius"
    elif simulator_name in ["riviera", "activehdl"]:
        library_name = "aldec"
    else:
        library_name = simulator_name

    if library_name == "icarus":
        lib_ext = ""
    elif os.name == "nt":
        lib_ext = ".dll"
    else:
        lib_ext = ".so"

    # check if compiled with msvc
    if (libs_dir / "cocotb.dll").is_file():
        lib_prefix = ""
    else:
        lib_prefix = "lib"

    return lib_prefix + "cocotb" + interface_name + "_" + library_name + lib_ext


def lib_name_path(interface: str, simulator: str) -> Path:
    """
    Return the absolute path of interface library for given interface (VPI/VHPI/FLI) and simulator
    """
    return libs_dir / lib_name(interface, simulator)


def _get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--share",
        action="store_true",
        help="Print the path to cocotb's share directory",
    )
    group.add_argument(
        "--makefiles",
        action="store_true",
        help="Print the path to cocotb's makefile directory",
    )
    group.add_argument(
        "--python-bin",
        action="store_true",
        help="Print the path to the Python executable associated with the environment that cocotb is installed in.",
    )
    group.add_argument(
        "--help-vars",
        action="store_true",
        help="Print help about supported Makefile variables",
    )
    group.add_argument(
        "--libpython",
        action="store_true",
        help="Print the absolute path to the libpython associated with the current Python installation",
    )
    group.add_argument(
        "--lib-dir",
        action="store_true",
        help="Print the absolute path to the interface libraries location",
    )
    group.add_argument(
        "--lib-name",
        help="Print the name of interface library for given interface (VPI/VHPI/FLI) and simulator",
        nargs=2,
        metavar=("INTERFACE", "SIMULATOR"),
    )
    group.add_argument(
        "--lib-name-path",
        help="Print the absolute path of interface library for given interface (VPI/VHPI/FLI) and simulator",
        nargs=2,
        metavar=("INTERFACE", "SIMULATOR"),
    )
    group.add_argument(
        "--version",
        action="store_true",
        help="Print the version of cocotb",
    )

    return parser


def main() -> None:
    parser = _get_parser()
    args = parser.parse_args()

    if args.share:
        print(share_dir.as_posix())
    elif args.makefiles:
        print(makefiles_dir.as_posix())
    elif args.python_bin:
        print(Path(sys.executable).as_posix())
    elif args.help_vars:
        print(_help_vars_text())
    elif args.libpython:
        libpython_path = find_libpython.find_libpython()
        if libpython_path is None:
            sys.exit(1)
        print(Path(libpython_path).as_posix())
    elif args.lib_dir:
        print(libs_dir.as_posix())
    elif args.lib_name:
        print(lib_name(*args.lib_name))
    elif args.lib_name_path:
        print(lib_name_path(*args.lib_name_path).as_posix())
    elif args.version:
        print(_get_version())


if __name__ == "__main__":
    main()



import importlib
import operator
import os
from typing import Callable, List, Tuple


def load_entry(argv: List[str]) -> None:
    """Gather entry point information by parsing :envvar:`PYGPI_USERS`."""

    entry_point_str = os.environ.get(
        "PYGPI_USERS",
        ",".join(
            (
                "cocotb_tools._coverage:start_cocotb_library_coverage",
                "cocotb.logging:_configure",
                "cocotb._init:init_package_from_simulation",
                "cocotb._init:run_regression",
            )
        ),
    )

    # Parse the entry point string of the form "module:func,module:func,...".
    # Any failure prevents any entry points from being loaded.
    entry_points: List[Tuple[str, str]] = []
    try:
        entry_points_str = entry_point_str.split(",")
        for entry_point_str in entry_points_str:
            entry_module_str, entry_func_str = entry_point_str.split(":")
            # TODO maybe some basic validation of the module and function names.
            # WITHOUT IMPORTING THEM.
            entry_points.append((entry_module_str, entry_func_str))
    except Exception as e:
        raise RuntimeError(f"Failure to parse PYGPI_USERS ('{entry_point_str}')") from e

    # Run all entry points.
    # Expect failure to stop the loading of any additional entry points.
    for entry_module_str, entry_func_str in entry_points:
        entry_module = importlib.import_module(entry_module_str)
        entry_func: Callable[[List[str]], object] = operator.attrgetter(entry_func_str)(
            entry_module
        )
        entry_func(argv)



# Copyright cocotb contributors
# Copyright (c) 2013, 2018 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""
Everything related to logging
"""

import logging
import os
import re
import sys
import time
import warnings
from functools import wraps
from typing import Optional, Tuple, Union

import cocotb.simtime
from cocotb import simulator
from cocotb._ANSI import ANSI
from cocotb._deprecation import deprecated
from cocotb.simtime import get_sim_time
from cocotb.utils import get_time_from_sim_steps

__all__ = (
    "ANSI",
    "SimLog",
    "SimLogFormatter",
    "SimTimeContextFilter",
    "default_config",
    "strip_ansi",
)

ANSI.__module__ = __name__

# Custom log level
logging.TRACE = 5  # type: ignore[attr-defined]  # type checkers don't like adding module attributes after the fact
logging.addLevelName(5, "TRACE")


strip_ansi: bool = False
"""Whether the default formatter should strip ANSI (color) escape codes from log messages.

Defaults to ``True`` if ``stdout`` is not a TTY and ``False`` otherwise;
but can be overridden with the :envvar:`NO_COLOR` or :envvar:`COCOTB_ANSI_OUTPUT` environment variable.
"""


def default_config(
    reduced_log_fmt: bool = True,
    strip_ansi: Optional[bool] = None,
    prefix_format: Optional[str] = None,
) -> None:
    """Apply the default cocotb log formatting to the root logger.

    This hooks up the logger to write to stdout, using :class:`SimLogFormatter` for formatting.
    It also adds a :class:`SimTimeContextFilter` filter so that the
    :attr:`~logging.LogRecord.created_sim_time` attribute on :class:`~logging.LogRecord`
    is available to the formatter.

    If desired, this logging configuration can be overwritten by calling
    ``logging.basicConfig(..., force=True)`` (in Python 3.8 onwards),
    or by manually resetting the root logger instance.
    An example of this can be found in the section on :ref:`rotating-logger`.

    Args:
        reduced_log_fmt:
            If ``True``, use a reduced log format that does not include the
            filename, line number, and function name in the log prefix.

            .. versionadded:: 2.0

        strip_ansi:
            If ``True``, strip ANSI (color) escape codes in log messages.
            If ``False``, do not strip ANSI escape codes in log messages.
            If ``None``, use the value of :data:`strip_ansi`.

            .. versionadded:: 2.0

        prefix_format:
            An f-string to build a prefix for each log message.

            .. versionadded:: 2.0

    .. versionadded:: 1.4

    .. versionchanged:: 2.0
        Now captures warnings and outputs them through the logging system using
        :func:`logging.captureWarnings`.
    """
    logging.basicConfig()

    hdlr = logging.StreamHandler(sys.stdout)
    hdlr.addFilter(SimTimeContextFilter())
    hdlr.setFormatter(
        SimLogFormatter(
            reduced_log_fmt=reduced_log_fmt,
            strip_ansi=strip_ansi,
            prefix_format=prefix_format,
        )
    )
    logging.getLogger().handlers = [hdlr]  # overwrite default handlers

    logging.getLogger("cocotb").setLevel(logging.INFO)
    logging.getLogger("gpi").setLevel(logging.INFO)

    logging.captureWarnings(True)


def _init() -> None:
    """cocotb-specific logging setup.

    - Decides whether ANSI escape code stripping is desired by checking
      :envvar:`NO_COLOR` and :envvar:`COCOTB_ANSI_OUTPUT`.
    - Initializes the GPI logger and sets up the GPI logging optimization.
    - Sets the log level of the ``"cocotb"`` and ``"gpi"`` loggers based on
      :envvar:`COCOTB_LOG_LEVEL` and :envvar:`GPI_LOG_LEVEL`, respectively.
    """
    global strip_ansi
    strip_ansi = not sys.stdout.isatty()  # default to color for TTYs
    if os.getenv("NO_COLOR", ""):
        strip_ansi = True
    ansi_output = os.getenv("COCOTB_ANSI_OUTPUT")
    if ansi_output is not None:
        strip_ansi = not int(ansi_output)
    in_gui = os.getenv("GUI")
    if in_gui is not None:
        strip_ansi = bool(int(in_gui))

    # Monkeypatch "gpi" logger with function that also sets a PyGPI-local logger level
    # as an optimization.
    gpi_logger = logging.getLogger("gpi")
    old_setLevel = gpi_logger.setLevel

    @wraps(old_setLevel)
    def setLevel(level: Union[int, str]) -> None:
        old_setLevel(level)
        simulator.set_gpi_log_level(gpi_logger.getEffectiveLevel())

    gpi_logger.setLevel = setLevel  # type: ignore[method-assign]

    # Initialize PyGPI logging
    simulator.initialize_logger(_log_from_c, logging.getLogger)

    # Set "cocotb" and "gpi" logger based on environment variables
    def set_level(logger_name: str, envvar: str) -> None:
        log_level = os.environ.get(envvar)
        if log_level is None:
            return

        log_level = log_level.upper()

        logger = logging.getLogger(logger_name)

        try:
            logger.setLevel(log_level)
        except ValueError:
            valid_levels = ", ".join(
                ("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "TRACE", "NOTSET")
            )
            raise ValueError(
                f"Invalid log level {log_level!r} passed through the "
                f"{envvar} environment variable. Valid log "
                f"levels: {valid_levels}"
            ) from None

    set_level("gpi", "GPI_LOG_LEVEL")
    set_level("cocotb", "COCOTB_LOG_LEVEL")


def _configure(_: object) -> None:
    """Configure basic logging."""
    reduced_log_fmt = True
    try:
        reduced_log_fmt = bool(int(os.environ.get("COCOTB_REDUCED_LOG_FMT", "1")))
    except ValueError:
        pass
    prefix_format = os.environ.get("COCOTB_LOG_PREFIX", None)
    default_config(reduced_log_fmt=reduced_log_fmt, prefix_format=prefix_format)


@deprecated('Use `logging.getLogger(f"{name}.0x{ident:x}")` instead')
def SimLog(name: str, ident: Union[int, None] = None) -> logging.Logger:
    """Like :func:`logging.getLogger`, but append a numeric identifier to the name.

    Args:
        name: Logger name.
        ident: Unique integer identifier.

    Returns:
        The Logger named ``{name}.0x{ident:x}``.

    .. deprecated:: 2.0

        Use ``logging.getLogger(f"{name}.0x{ident:x}")`` instead.
    """
    if ident is not None:
        name = f"{name}.0x{ident:x}"
    return logging.getLogger(name)


class SimTimeContextFilter(logging.Filter):
    """
    A filter to inject simulator times into the log records.

    This uses the approach described in the :ref:`Python logging cookbook <python:filters-contextual>`
    which adds the :attr:`~logging.LogRecord.created_sim_time` attribute.

    .. versionadded:: 1.4
    """

    def filter(self, record: logging.LogRecord) -> bool:
        try:
            record.created_sim_time = get_sim_time()
        except RecursionError:
            # get_sim_time may try to log - if that happens, we can't
            # attach a simulator time to this message.
            record.created_sim_time = None
        return True


class SimLogFormatter(logging.Formatter):
    """Log formatter to provide consistent log message handling.

    This will only add simulator timestamps if the handler object this
    formatter is attached to has a :class:`SimTimeContextFilter` filter
    attached, which cocotb ensures by default.

    See :func:`.default_config` for a description of the arguments.
    """

    loglevel2colour = {
        logging.TRACE: "",  # type: ignore[attr-defined]  # type checkers don't like adding module attributes after the fact
        logging.DEBUG: "",
        logging.INFO: "",
        logging.WARNING: ANSI.YELLOW_FG,
        logging.ERROR: ANSI.RED_FG,
        logging.CRITICAL: ANSI.RED_BG + ANSI.BLACK_FG,
    }

    prefix_func_globals = {
        "time": time,
        "simtime": cocotb.simtime,
        "ANSI": ANSI,
    }

    def __init__(
        self,
        *,
        reduced_log_fmt: bool = True,
        strip_ansi: Union[bool, None] = None,
        prefix_format: Optional[str] = None,
    ) -> None:
        self._reduced_log_fmt = reduced_log_fmt
        self._strip_ansi = strip_ansi
        self._prefix_func = (
            eval(
                f"lambda record: f'''{prefix_format}'''",
                type(self).prefix_func_globals,
            )
            if prefix_format is not None
            else None
        )
        self._ansi_escape_pattern = re.compile(
            r"""
                \x1B
                (?: # either 7-bit C1, two bytes, ESC Fe (omitting CSI)
                    [@-Z\\-_]
                | # or 7-bit CSI (ESC [) + control codes
                    \[
                    [0-?]*  # Parameter bytes
                    [ -/]*  # Intermediate bytes
                    [@-~]   # Final byte
                )
            """,
            re.VERBOSE,
        )

    def strip_ansi(self) -> bool:
        return strip_ansi if self._strip_ansi is None else self._strip_ansi

    # Justify and truncate
    @staticmethod
    def ljust(string: str, chars: int) -> str:
        if len(string) > chars:
            return ".." + string[(chars - 2) * -1 :]
        return string.ljust(chars)

    @staticmethod
    def rjust(string: str, chars: int) -> str:
        if len(string) > chars:
            return ".." + string[(chars - 2) * -1 :]
        return string.rjust(chars)

    def formatPrefix(self, record: logging.LogRecord) -> Tuple[str, int]:
        sim_time = getattr(record, "created_sim_time", None)
        if sim_time is None:
            sim_time_str = "  -.--ns"
        else:
            time_ns = get_time_from_sim_steps(sim_time, "ns")
            sim_time_str = f"{time_ns:6.2f}ns"

        if self.strip_ansi():
            highlight_start = ""
            highlight_end = ""
        else:
            highlight_start = self.loglevel2colour.get(record.levelno, "")
            highlight_end = ANSI.DEFAULT

        if self._prefix_func is not None:
            prefix = self._prefix_func(record)
            stripped_prefix = self._ansi_escape_pattern.sub("", prefix)
            prefix_len = len(stripped_prefix)
            return prefix, prefix_len
        else:
            prefix = f"{sim_time_str:>11} {highlight_start}{record.levelname:<8}{highlight_end} {self.ljust(record.name, 34)} "
            if not self._reduced_log_fmt:
                prefix = f"{prefix}{self.rjust(record.filename, 20)}:{record.lineno:<4} in {self.ljust(str(record.funcName), 31)} "

            prefix_len = len(prefix) - len(highlight_start) - len(highlight_end)
            return prefix, prefix_len

    def formatMessage(self, record: logging.LogRecord) -> str:
        msg = record.getMessage()

        if self.strip_ansi():
            highlight_start = ""
            highlight_end = ""
            msg = self._ansi_escape_pattern.sub("", msg)
        else:
            highlight_start = self.loglevel2colour.get(record.levelno, "")
            highlight_end = ANSI.DEFAULT

        msg = f"{highlight_start}{msg}{highlight_end}"

        return msg

    def formatExcInfo(self, record: logging.LogRecord) -> str:
        msg = ""

        # these lines are copied and updated from the built-in logger
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            msg += record.exc_text
        if record.stack_info:
            if not msg.endswith("\n"):
                msg += "\n"
            msg += self.formatStack(record.stack_info)

        return msg

    def format(self, record: logging.LogRecord) -> str:
        prefix, prefix_len = self.formatPrefix(record)
        msg = self.formatMessage(record)
        exc_info = self.formatExcInfo(record)

        lines = msg.splitlines()
        lines.extend(exc_info.splitlines())

        # add prefix to first line of message
        lines[0] = prefix + lines[0]

        # add padding to each line of message
        pad = "\n" + " " * prefix_len
        return pad.join(lines)


class SimColourLogFormatter(SimLogFormatter):
    """Log formatter similar to :class:`SimLogFormatter`, but with colored output by default.

    .. deprecated:: 2.0
        Use :class:`!SimLogFormatter` with ``strip_ansi=False`` instead.
    """

    def __init__(
        self,
        *,
        reduced_log_fmt: bool = True,
        strip_ansi: Union[bool, None] = False,
        prefix_format: Optional[str] = None,
    ) -> None:
        warnings.warn(
            "SimColourLogFormatter is deprecated and will be removed in a future release. "
            "Use SimLogFormatter with `strip_ansi=False` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(
            reduced_log_fmt=reduced_log_fmt,
            strip_ansi=strip_ansi,
            prefix_format=prefix_format,
        )


def _log_from_c(
    logger: logging.Logger,
    level: int,
    filename: str,
    lineno: int,
    msg: str,
    function_name: str,
) -> None:
    """
    This is for use from the C world, and allows us to insert C stack
    information.
    """
    if logger.isEnabledFor(level):
        record = logger.makeRecord(
            logger.name, level, filename, lineno, msg, (), None, function_name
        )
        logger.handle(record)



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import asyncio.queues
import collections
import heapq
from abc import abstractmethod
from typing import TYPE_CHECKING, Deque, Generic, List, Tuple, TypeVar

import cocotb
from cocotb._utils import pointer_str
from cocotb.task import Task
from cocotb.triggers import Event

if TYPE_CHECKING:
    from typing import Protocol, Self

    class SupportsRichComparison(Protocol):
        def __eq__(self, other: object) -> bool: ...
        def __lt__(self, other: Self) -> bool: ...
        def __le__(self, other: Self) -> bool: ...
        def __gt__(self, other: Self) -> bool: ...
        def __ge__(self, other: Self) -> bool: ...


__all__ = (
    "AbstractQueue",
    "LifoQueue",
    "PriorityQueue",
    "Queue",
    "QueueEmpty",
    "QueueFull",
)


class QueueFull(asyncio.queues.QueueFull):
    """Raised when the :meth:`Queue.put_nowait()` method is called on a full :class:`Queue`."""


class QueueEmpty(asyncio.queues.QueueEmpty):
    """Raised when the :meth:`Queue.get_nowait()` method is called on an empty :class:`Queue`."""


T = TypeVar("T")


class AbstractQueue(Generic[T]):
    """A queue, useful for coordinating producer and consumer coroutines.

    If *maxsize* is less than or equal to 0, the queue size is infinite. If it
    is an integer greater than 0, then :meth:`put` will block when the queue
    reaches *maxsize*, until an item is removed by :meth:`get`.
    """

    def __init__(self, maxsize: int = 0) -> None:
        self._maxsize: int = maxsize
        self._getters: Deque[Tuple[Event, Task[object]]] = collections.deque()
        self._putters: Deque[Tuple[Event, Task[object]]] = collections.deque()

    @abstractmethod
    def _get(self) -> T:
        """Remove and return the next element from the queue."""

    @abstractmethod
    def _put(self, item: T) -> None:
        """Place a new element on the queue."""

    @abstractmethod
    def _size(self) -> int:
        """Return the number of elements in the queue."""

    @abstractmethod
    def _repr(self) -> str:
        """Return a string representation of the state of the queue."""

    def _wakeup_next(self, waiters: Deque[Tuple[Event, Task[object]]]) -> None:
        while waiters:
            event, task = waiters.popleft()
            if not task.done():
                event.set()
                break

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self._format()} at {pointer_str(self)}>"

    def __str__(self) -> str:
        return f"<{type(self).__name__} {self._format()}>"

    def _format(self) -> str:
        result = f"maxsize={self._maxsize!r}"
        if getattr(self, "_queue", None):
            result += f" _queue={self._repr()}"
        if self._getters:
            result += f" _getters[{len(self._getters)}]"
        if self._putters:
            result += f" _putters[{len(self._putters)}]"
        return result

    def qsize(self) -> int:
        """Number of items in the queue."""
        return self._size()

    @property
    def maxsize(self) -> int:
        """Number of items allowed in the queue."""
        return self._maxsize

    def empty(self) -> bool:
        """Return ``True`` if the queue is empty, ``False`` otherwise."""
        return self._size() == 0

    def full(self) -> bool:
        """Return ``True`` if there are :meth:`maxsize` items in the queue.

        .. note::
            If the Queue was initialized with ``maxsize=0`` (the default), then
            :meth:`full` is never ``True``.
        """
        if self._maxsize <= 0:
            return False
        else:
            return self.qsize() >= self._maxsize

    async def put(self, item: T) -> None:
        """Put an *item* into the queue.

        If the queue is full, wait until a free
        slot is available before adding the item.
        """
        while self.full():
            event = Event()
            self._putters.append((event, cocotb.task.current_task()))
            await event.wait()
        self.put_nowait(item)

    def put_nowait(self, item: T) -> None:
        """Put an *item* into the queue without blocking.

        If no free slot is immediately available, raise :exc:`~cocotb.queue.QueueFull`.
        """
        if self.full():
            raise QueueFull()
        self._put(item)
        self._wakeup_next(self._getters)

    async def get(self) -> T:
        """Remove and return an item from the queue.

        If the queue is empty, wait until an item is available.
        """
        while self.empty():
            event = Event()
            self._getters.append((event, cocotb.task.current_task()))
            await event.wait()
        return self.get_nowait()

    def get_nowait(self) -> T:
        """Remove and return an item from the queue.

        Return an item if one is immediately available, else raise
        :exc:`~cocotb.queue.QueueEmpty`.
        """
        if self.empty():
            raise QueueEmpty()
        item = self._get()
        self._wakeup_next(self._putters)
        return item


class Queue(AbstractQueue[T]):
    """A subclass of :class:`AbstractQueue`; retrieves oldest entries first (FIFO)."""

    def __init__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize)
        self._queue: Deque[T] = collections.deque()

    def _put(self, item: T) -> None:
        self._queue.append(item)

    def _get(self) -> T:
        return self._queue.popleft()

    def _size(self) -> int:
        return len(self._queue)

    def _repr(self) -> str:
        return repr(self._queue)


SupportsRichComparisonT = TypeVar(
    "SupportsRichComparisonT", bound="SupportsRichComparison"
)


class PriorityQueue(AbstractQueue[SupportsRichComparisonT]):
    r"""A subclass of :class:`AbstractQueue`; retrieves entries in priority order (smallest item first).

    Entries are typically tuples of the form ``(priority number, data)``.
    """

    def __init__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize)
        self._queue: List[SupportsRichComparisonT] = []

    def _put(self, item: SupportsRichComparisonT) -> None:
        heapq.heappush(self._queue, item)

    def _get(self) -> SupportsRichComparisonT:
        return heapq.heappop(self._queue)

    def _size(self) -> int:
        return len(self._queue)

    def _repr(self) -> str:
        return repr(self._queue)


class LifoQueue(AbstractQueue[T]):
    """A subclass of :class:`AbstractQueue`; retrieves most recently added entries first (LIFO)."""

    def __init__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize)
        self._queue: Deque[T] = collections.deque()

    def _put(self, item: T) -> None:
        self._queue.append(item)

    def _get(self) -> T:
        return self._queue.pop()

    def _size(self) -> int:
        return len(self._queue)

    def _repr(self) -> str:
        return repr(self._queue)



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""A collection of triggers which a testbench can :keyword:`await`."""

from abc import abstractmethod
from decimal import Decimal
from typing import (
    Any,
    Awaitable,
    Coroutine,
    Generator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import cocotb.handle
from cocotb._base_triggers import NullTrigger, Trigger, _InternalEvent
from cocotb._gpi_triggers import FallingEdge, RisingEdge, Timer, ValueChange
from cocotb._typing import RoundMode, TimeUnit
from cocotb.task import Task

T = TypeVar("T")


class Waitable(Awaitable[T]):
    """A Trigger-like object that can be implemented using coroutines.

    This converts a ``_wait`` abstract method into a suitable ``__await__``.
    """

    @abstractmethod
    async def _wait(self) -> T:
        """The coroutine function which implements the functionality of the Waitable."""

    def __await__(self) -> Generator[Trigger, None, T]:
        return self._wait().__await__()


class _AggregateWaitable(Waitable[T]):
    """Base class for :class:`Combine` and :class:`First`."""

    def __init__(self, *trigger: Union[Trigger, Waitable[Any], Task[Any]]) -> None:
        self._triggers = trigger

        # Do some basic type-checking up front, rather than waiting until we
        # await them.
        allowed_types = (Trigger, Waitable, Task)
        for t in self._triggers:
            if not isinstance(t, allowed_types):
                raise TypeError(
                    f"All triggers must be instances of Trigger! Got: {type(t).__qualname__}"
                )

    def __repr__(self) -> str:
        # no _pointer_str here, since this is not a trigger, so identity
        # doesn't matter.
        return "{}({})".format(
            type(self).__qualname__,
            ", ".join(repr(t) for t in self._triggers),
        )


async def _wait_callback(trigger: Awaitable[T]) -> T:
    return await trigger


class Combine(_AggregateWaitable["Combine"]):
    r"""Trigger that fires when all *triggers* have fired.

    :keyword:`await`\ ing this returns the :class:`Combine` object.
    This is similar to Verilog's ``join``.
    See :ref:`combine-tutorial` for an example.

    Args:
        trigger: One or more :keyword:`await`\ able objects.

    Raises:
        TypeError: When an unsupported *trigger* object is passed.
    """

    async def _wait(self) -> "Combine":
        if len(self._triggers) == 0:
            await NullTrigger()
        elif len(self._triggers) == 1:
            await self._triggers[0]
        else:
            waiters: List[Task[object]] = []
            completed: List[Task[object]] = []
            done = _InternalEvent(self)
            exception: Union[BaseException, None] = None

            def on_done(
                task: Task[object],
            ) -> None:
                # have to check cancelled first otherwise exception() will throw
                if task.cancelled():
                    completed.append(task)
                    if len(completed) == len(waiters):
                        done.set()
                    return
                e = task.exception()
                if e is not None:
                    nonlocal exception
                    exception = e
                    done.set()
                else:
                    completed.append(task)
                    if len(completed) == len(waiters):
                        done.set()

            # start a parallel task for each trigger
            for t in self._triggers:
                task = Task[object](_wait_callback(t))
                task._add_done_callback(on_done)
                cocotb.start_soon(task)
                waiters.append(task)

            try:
                # wait for the last waiter to complete
                await done
            finally:
                # kill remaining waiters
                for w in waiters:
                    w.cancel()

            if exception is not None:
                raise exception

        return self


class First(_AggregateWaitable[object]):
    r"""Fires when the first trigger in *triggers* fires.

    :keyword:`await`\ ing this object returns the result of the first trigger that fires.
    This is similar to Verilog's ``join_any``.
    See :ref:`first-tutorial` for an example.

    Args:
        trigger: One or more :keyword:`await`\ able objects.

    Raises:
        TypeError: When an unsupported *trigger* object is passed.
        ValueError: When no triggers are passed.

    .. note::
        The event loop is single threaded, so while events may be simultaneous
        in simulation time, they can never be simultaneous in real time.
        For this reason, the value of ``t_ret is t1`` in the following example
        is implementation-defined, and will vary by simulator::

            t1 = Timer(10, unit="ps")
            t2 = Timer(10, unit="ps")
            t_ret = await First(t1, t2)

    .. note::
        In the old-style :ref:`generator-based coroutines <yield-syntax>`, ``t = yield [a, b]`` was another spelling of
        ``t = yield First(a, b)``. This spelling is no longer available when using :keyword:`await`-based
        coroutines.
    """

    def __init__(self, *trigger: Union[Trigger, Waitable[Any], Task[Any]]) -> None:
        if not trigger:
            raise ValueError("First() requires at least one Trigger or Task argument")
        super().__init__(*trigger)

    async def _wait(self) -> object:
        if len(self._triggers) == 1:
            return await self._triggers[0]

        waiters: List[Task[object]] = []
        done = _InternalEvent(self)
        completed: List[Task[object]] = []

        def on_done(task: Task[object]) -> None:
            completed.append(task)
            done.set()

        # start a parallel task for each trigger
        for t in self._triggers:
            task = Task[object](_wait_callback(t))
            task._add_done_callback(on_done)
            cocotb.start_soon(task)
            waiters.append(task)

        try:
            # wait for a waiter to complete
            await done
        finally:
            # kill all the other waiters
            for w in waiters:
                w.cancel()

        return completed[0].result()


class ClockCycles(Waitable["ClockCycles"]):
    r"""Finishes after *num_cycles* transitions of *signal*.

    :keyword:`await`\ ing this Trigger returns the :class:`!ClockCycles` object.

    Args:
        signal: The signal to monitor.
        num_cycles: The number of cycles to count.
        rising: If ``True``, count rising edges; if ``False``, count falling edges.
        edge_type: The kind of :ref:`edge-triggers` to count.

    .. warning::
        On many simulators transitions occur when the signal changes value from non-``0`` to ``0`` or non-``1`` to ``1``,
        not just from ``1`` to ``0`` or ``0`` to ``1``.

    .. versionadded:: 2.0
        Passing the edge trigger type: :class:`.RisingEdge`, :class:`.FallingEdge`, or :class:`.ValueChange`
        as the third positional argument or by the keyword *edge_type*.
    """

    @overload
    def __init__(
        self,
        signal: "cocotb.handle.LogicObject",
        num_cycles: int,
    ) -> None: ...

    @overload
    def __init__(
        self,
        signal: "cocotb.handle.LogicObject",
        num_cycles: int,
        edge_type: Union[
            Type[RisingEdge], Type[FallingEdge], Type[ValueChange], None
        ] = None,
    ) -> None: ...

    @overload
    def __init__(
        self, signal: "cocotb.handle.LogicObject", num_cycles: int, *, rising: bool
    ) -> None: ...

    def __init__(
        self,
        signal: "cocotb.handle.LogicObject",
        num_cycles: int,
        edge_type: Union[
            bool, Type[RisingEdge], Type[FallingEdge], Type[ValueChange], None
        ] = None,
        *,
        rising: Union[bool, None] = None,
    ) -> None:
        self._signal = signal
        self._num_cycles = num_cycles
        self._edge_type: Union[Type[RisingEdge], Type[FallingEdge], Type[ValueChange]]
        if edge_type is not None and rising is not None:
            raise TypeError("Passed more than one edge selection argument.")
        elif edge_type is True:
            self._edge_type = RisingEdge
        elif edge_type is False:
            self._edge_type = FallingEdge
        elif edge_type is not None:
            self._edge_type = edge_type
        elif rising is not None:
            self._edge_type = RisingEdge if rising else FallingEdge
        else:
            # default if no argument is passed
            self._edge_type = RisingEdge

    @property
    def signal(self) -> "cocotb.handle.LogicObject":
        """The signal being monitored."""
        return self._signal

    @property
    def num_cycles(self) -> int:
        """The number of cycles to wait."""
        return self._num_cycles

    @property
    def edge_type(
        self,
    ) -> Union[Type[RisingEdge], Type[FallingEdge], Type[ValueChange]]:
        """The type of edge trigger used."""
        return self._edge_type

    async def _wait(self) -> "ClockCycles":
        trigger = self._edge_type(self._signal)
        for _ in range(self._num_cycles):
            await trigger
        return self

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}({self._signal._path}, {self._num_cycles}, {self._edge_type.__qualname__})"


class SimTimeoutError(TimeoutError):
    """Exception thrown when a timeout, in terms of simulation time, occurs."""


TriggerT = TypeVar("TriggerT", bound=Trigger)


@overload
async def with_timeout(
    trigger: TriggerT,
    timeout_time: Union[float, Decimal],
    timeout_unit: TimeUnit = "step",
    round_mode: Optional[RoundMode] = None,
) -> TriggerT: ...


@overload
async def with_timeout(
    trigger: Waitable[T],
    timeout_time: Union[float, Decimal],
    timeout_unit: TimeUnit = "step",
    round_mode: Optional[RoundMode] = None,
) -> T: ...


@overload
async def with_timeout(
    trigger: Task[T],
    timeout_time: Union[float, Decimal],
    timeout_unit: TimeUnit = "step",
    round_mode: Optional[RoundMode] = None,
) -> T: ...


@overload
async def with_timeout(
    trigger: Coroutine[Trigger, None, T],
    timeout_time: Union[float, Decimal],
    timeout_unit: TimeUnit = "step",
    round_mode: Optional[RoundMode] = None,
) -> T: ...


async def with_timeout(
    trigger: Union[TriggerT, Waitable[T], Task[T], Coroutine[Trigger, None, T]],
    timeout_time: Union[float, Decimal],
    timeout_unit: TimeUnit = "step",
    round_mode: Optional[RoundMode] = None,
) -> Union[T, TriggerT]:
    r"""Wait on triggers or coroutines, throw an exception if it waits longer than the given time.

    When a :term:`python:coroutine` is passed,
    the callee coroutine is started,
    the caller blocks until the callee completes,
    and the callee's result is returned to the caller.
    If timeout occurs, the callee is killed
    and :exc:`SimTimeoutError` is raised.

    When a :term:`task` is passed,
    the caller blocks until the callee completes
    and the callee's result is returned to the caller.
    If timeout occurs, the callee `continues to run`
    and :exc:`SimTimeoutError` is raised.

    If a :class:`~cocotb.triggers.Trigger` or :class:`~cocotb.triggers.Waitable` is passed,
    the caller blocks until the trigger fires,
    and the trigger is returned to the caller.
    If timeout occurs, the trigger is cancelled
    and :exc:`SimTimeoutError` is raised.

    Usage:
        .. code-block:: python

            await with_timeout(coro, 100, "ns")
            await with_timeout(First(coro, event.wait()), 100, "ns")

    Args:
        trigger:
            A single object that could be right of an :keyword:`await` expression in cocotb.
        timeout_time:
            Simulation time duration before timeout occurs.
        timeout_unit:
            Unit of timeout_time, accepts any unit that :class:`~cocotb.triggers.Timer` does.
        round_mode:
            String specifying how to handle time values that sit between time steps
            (one of ``'error'``, ``'round'``, ``'ceil'``, ``'floor'``, ``None``).
            A ``None`` argument is converted to the current value of :attr:`.Timer.round_mode`.

    Returns:
        First trigger that completed if timeout did not occur.

    Raises:
        :exc:`SimTimeoutError`: If timeout occurs.

    .. versionadded:: 1.3

    .. versionchanged:: 1.7
        Support passing :term:`python:coroutine`\ s.

    .. versionchanged:: 2.0
        Passing ``None`` as the *timeout_unit* argument was removed, use ``'step'`` instead.

    """
    if isinstance(trigger, Coroutine):
        trigger = cocotb.start_soon(trigger)
        shielded = False
    else:
        shielded = True
    timeout_timer = Timer(timeout_time, timeout_unit, round_mode=round_mode)
    res = await First(timeout_timer, trigger)
    if res is timeout_timer:
        if not shielded:
            # shielded = False only when trigger is a Task created to wrap a Coroutine
            task = cast("Task[object]", trigger)
            task.cancel()
        raise SimTimeoutError
    else:
        return cast("T | TriggerT", res)



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

import os

from cocotb.types._range import Range

do_indexing_changed_warning = os.environ.get("COCOTB_INDEXING_CHANGED_WARNING")


def indexing_changed(range: Range) -> bool:
    return not (range.left == 0 and range.direction == "to")


class IndexingChangedWarning(UserWarning):
    """Warning issued when a value is indexed in a way that is different between cocotb 1.x and 2.x."""
